# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for base workflow code."""

import logging
from functools import partial
from typing import ClassVar, override
from unittest import mock

import debusine.tasks.tags as ttags
from debusine.artifacts.models import WorkRequestResults
from debusine.client.models import RuntimeParameter
from debusine.db.context import context
from debusine.db.models import WorkRequest
from debusine.db.playground import scenarios
from debusine.server.tasks.wait.models import ExternalDebsignData
from debusine.server.workflows.base import (
    Workflow,
    WorkflowValidationError,
    orchestrate_workflow,
)
from debusine.server.workflows.models import (
    BaseWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import SampleWorkflow
from debusine.tasks.models import (
    BaseDynamicTaskData,
    OutputData,
    OutputDataError,
)
from debusine.test.django import TestCase, preserve_db_task_registry


class WorkflowTests(TestCase):
    """Test :py:class:`Workflow`."""

    @preserve_db_task_registry()
    def make_concrete_workflow(
        self,
    ) -> Workflow[BaseWorkflowData, BaseDynamicTaskData]:
        """Instantiate a concrete instance of Workflow."""

        class ConcreteWorkflow(Workflow[BaseWorkflowData, BaseDynamicTaskData]):
            """Concrete implementation of Workflow."""

            TASK_VERSION = 1

            @override
            def populate(self) -> None:
                raise NotImplementedError()

            @override
            def compute_dynamic_data(self) -> BaseDynamicTaskData:
                return BaseDynamicTaskData()

            @override
            def _execute(self) -> WorkRequestResults:
                raise NotImplementedError()

        work_request = self.playground.create_workflow(
            task_name="concreteworkflow"
        )
        task = work_request.get_task()
        assert isinstance(task, Workflow)
        return task

    def test_get_label_no_workflow_data_no_dynamic_data(self) -> None:
        task = self.make_concrete_workflow()
        task.work_request.workflow_data_json = {}
        self.assertEqual(task.get_label(), "concreteworkflow")

    def test_get_label_workflow_data_no_dynamic_data(self) -> None:
        task = self.make_concrete_workflow()
        task.work_request.workflow_data_json = {
            "workflow_template_name": "concreteworkflow-template"
        }
        self.assertEqual(task.get_label(), "concreteworkflow-template")

    def test_get_label_no_parameter_summary(self) -> None:
        task = self.make_concrete_workflow()
        task.ensure_dynamic_data()
        self.assertEqual(task.get_label(), "concreteworkflow")

    def test_get_label_parameter_summary(self) -> None:
        task = self.make_concrete_workflow()
        task.ensure_dynamic_data()
        assert task.dynamic_data is not None
        task.dynamic_data.parameter_summary = "summary"
        self.assertEqual(task.get_label(), "concreteworkflow: summary")

    def test_get_label_callback_no_wf_data_no_dyn_data(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        workflow.workflow_data_json = {}
        workflow.save()
        wr = WorkRequest.objects.create_workflow_callback(
            parent=workflow, step="test", display_name="Test"
        )
        wr.workflow_data_json = {}
        wr.save()
        self.assertEqual(wr.get_task().get_label(), "noop")

    def test_get_label_callback_no_dyn_data_no_display_name(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        wr = WorkRequest.objects.create_workflow_callback(
            parent=workflow, step="test-step"
        )
        self.assertEqual(wr.get_task().get_label(), "test-step")

    def test_get_label_callback_no_dyn_data(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        wr = WorkRequest.objects.create_workflow_callback(
            parent=workflow, step="test-step", display_name="Test Step"
        )
        self.assertEqual(wr.get_task().get_label(), "Test Step")

    def test_get_label_callback_parameter_summary(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        parent = self.playground.create_workflow(template, task_data={})
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test-step", display_name="Test Step"
        )
        workflow = wr.get_task()
        assert isinstance(workflow, Workflow)
        workflow.ensure_dynamic_data()
        assert workflow.dynamic_data is not None
        workflow.dynamic_data.parameter_summary = "summary"
        self.assertEqual(workflow.get_label(), "Test Step: summary")

    def test_get_label_workflow_template_name(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        workflow.workflow_data_json["workflow_template_name"] = "template-name"
        workflow.save()

        self.assertEqual(
            workflow.get_task().get_label(),
            "template-name",
        )

    def test_get_label_workflow_include_parameter_summary(self) -> None:
        template = self.playground.create_workflow_template(
            name="test", task_name="noop"
        )
        workflow = self.playground.create_workflow(template, task_data={})
        workflow.dynamic_task_data = {"parameter_summary": "hello/amd64"}
        workflow.save()

        self.assertEqual(
            workflow.get_task().get_label(),
            "test: hello/amd64",
        )

    def test_get_description(self) -> None:
        task = self.make_concrete_workflow()
        self.assertEqual(task.get_description(), task.__doc__)

    def test_get_description_callback(self) -> None:
        def cb_function() -> None:
            """Test documentation."""
            raise NotImplementedError()

        callback = self.playground.create_internal_task(task_name="workflow")
        callback.workflow_data_json = {"step": "test"}
        task = self.make_concrete_workflow()
        task.work_request_callback = callback
        task.callback_test = cb_function  # type: ignore[attr-defined]
        assert task.__doc__ is not None
        assert cb_function.__doc__ is not None
        self.assertEqual(task.get_description(), cb_function.__doc__)

    def test_get_description_callback_no_step(self) -> None:
        callback = self.playground.create_internal_task(task_name="workflow")
        task = self.make_concrete_workflow()
        task.work_request_callback = callback
        self.assertEqual(task.get_description(), task.__doc__)

    def test_get_description_callback_invalid_step(self) -> None:
        callback = self.playground.create_internal_task(task_name="workflow")
        callback.workflow_data_json = {"step": "test"}
        task = self.make_concrete_workflow()
        task.work_request_callback = callback
        self.assertEqual(task.get_description(), task.__doc__)

    def test_get_description_callback_step_without_docstring(self) -> None:
        def cb_function() -> None:
            raise NotImplementedError()

        callback = self.playground.create_internal_task(task_name="workflow")
        callback.workflow_data_json = {"step": "test"}
        task = self.make_concrete_workflow()
        task.work_request_callback = callback
        task.callback_test = cb_function  # type: ignore[attr-defined]

        assert task.__doc__ is not None
        self.assertEqual(task.get_description(), task.__doc__)

    def test_compute_scheduler_tags_no_source_package_name(self) -> None:
        task = self.make_concrete_workflow()
        early_dynamic_task_data = mock.Mock()
        early_dynamic_task_data.get_source_package_name = mock.Mock(
            return_value=None
        )
        with mock.patch.object(
            task, "compute_dynamic_data", return_value=early_dynamic_task_data
        ):
            provided, _ = task.compute_scheduler_tags()
        self.assertCountEqual(
            provided.sets[ttags.ProvenanceProvided.USER], set()
        )

    def test_compute_scheduler_tags_with_source_package_name(self) -> None:
        task = self.make_concrete_workflow()
        early_dynamic_task_data = mock.Mock()
        early_dynamic_task_data.get_source_package_name = mock.Mock(
            return_value="foo"
        )
        with mock.patch.object(
            task, "compute_dynamic_data", return_value=early_dynamic_task_data
        ):
            provided, _ = task.compute_scheduler_tags()
        self.assertCountEqual(
            provided.sets[ttags.ProvenanceProvided.SYSTEM],
            {
                ttags.SOURCE_PACKAGE_PREFIX + "foo",
                ttags.SCOPE_PREFIX + "debusine",
                ttags.WORKSPACE_PREFIX + "debusine:System",
            },
        )


class OrchestrateWorkflowTests(TestCase):
    """Test orchestrate_workflow()."""

    def test_workflow_callback_success(self) -> None:
        """A workflow callback is run and marked as completed."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=True,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow_callback_failure(self) -> None:
        """A workflow callback that returns False is marked as failed."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=False,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.FAILURE)

    def test_workflow_callback_incremental_unterminated_dependencies(
        self,
    ) -> None:
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent,
            unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
            step="test",
        )
        dep = parent.create_child_worker("noop")
        wr.add_dependency(dep)
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=True,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)

    def test_workflow_callback_incremental_no_unterminated_dependencies(
        self,
    ) -> None:
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent,
            unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
            step="test",
        )
        dep = parent.create_child_worker("noop")
        self.playground.advance_work_request(
            dep, result=WorkRequest.Results.SUCCESS
        )
        wr.add_dependency(dep)
        wr.mark_pending()

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.callback_test",
            create=True,
            return_value=True,
        ) as mock_noop_callback_test:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_callback_test.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow_callback_step_not_implemented(self) -> None:
        """A workflow callback must have a matching ``callback_*`` method."""
        parent = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = (
            "Orchestrator failed: Unhandled workflow callback step: test"
        )

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback_other",
                create=True,
                return_value=True,
            ) as mock_noop_callback_other,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback_other.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    @preserve_db_task_registry()
    def test_workflow_callback_computes_dynamic_data(self) -> None:
        """Dynamic task data is computed for workflow callbacks."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            @override
            def compute_dynamic_data(self) -> ExampleDynamicData:
                return ExampleDynamicData(dynamic="foo")

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        parent = self.playground.create_workflow(
            task_name="example", mark_running=True
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with mock.patch.object(ExampleWorkflow, "callback") as mock_callback:
            self.assertTrue(orchestrate_workflow(wr))

        mock_callback.assert_called_once_with(wr)
        parent.refresh_from_db()
        self.assertEqual(parent.dynamic_task_data, {"dynamic": "foo"})
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    @preserve_db_task_registry()
    def test_workflow_callback_computes_dynamic_data_only_once(self) -> None:
        """If a workflow already has dynamic data, it is left alone."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str = "unset"

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            @override
            def populate(self) -> None:
                raise NotImplementedError()

        WorkRequest.objects.all().delete()
        parent = self.playground.create_workflow(task_name="example")
        parent.dynamic_task_data = {"dynamic": "foo"}
        parent.save()
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()

        with (
            mock.patch.object(
                ExampleWorkflow, "compute_dynamic_data"
            ) as mock_compute_dynamic_data,
            mock.patch.object(ExampleWorkflow, "callback") as mock_callback,
        ):
            self.assertTrue(orchestrate_workflow(wr))

        mock_compute_dynamic_data.assert_not_called()
        mock_callback.assert_called_once_with(wr)
        parent.refresh_from_db()
        self.assertEqual(parent.dynamic_task_data, {"dynamic": "foo"})
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    def test_workflow_callback_outside_workflow(self) -> None:
        """A workflow callback outside a workflow is skipped."""
        wr = self.playground.create_internal_task(task_name="workflow")
        self.playground.advance_work_request(
            wr, mark_pending=True, use_data=False
        )
        expected_message = "Workflow callback is not contained in a workflow"

        with self.assertLogsContains(
            f"Error running work request Internal/workflow ({wr.id}): "
            f"{expected_message}",
            logger="debusine.server.workflows.base",
            level=logging.WARNING,
        ):
            self.assertFalse(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="configure-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_bad_task_data(self) -> None:
        """A workflow callback with bad task data is skipped."""
        parent = self.playground.create_workflow(
            task_name="noop", task_data={"nonsense": ""}, validate=False
        )
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Failed to configure: 1 validation error"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback"
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        assert isinstance(wr.output_data, OutputData)
        assert wr.output_data.errors is not None
        self.assertEqual(len(wr.output_data.errors), 1)
        self.assertTrue(
            wr.output_data.errors[0].message.startswith(expected_message)
        )
        self.assertEqual(wr.output_data.errors[0].code, "configure-failed")

    def test_workflow_callback_with_failing_compute_dynamic_data(self) -> None:
        """A workflow callback where computing dynamic data fails is skipped."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = (
            f"Failed to compute dynamic data for Workflow/noop ({parent.id}): "
            f"Boom"
        )

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow"
                ".compute_dynamic_data",
                side_effect=Exception("Boom"),
            ),
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback"
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="dynamic-data-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_fails(self) -> None:
        """A workflow callback that fails is logged."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Orchestrator failed: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback",
                side_effect=ValueError("Boom"),
            ) as mock_noop_callback,
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_callback.assert_called_once_with(wr)
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow_callback_fails_without_context(self) -> None:
        """A callback failure by bare exception logs the exception name."""
        parent = self.playground.create_workflow(task_name="noop")
        wr = WorkRequest.objects.create_workflow_callback(
            parent=parent, step="test"
        )
        wr.mark_pending()
        expected_message = "Orchestrator failed: AssertionError()"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.callback",
                side_effect=AssertionError(),
            ),
            self.assertLogsContains(
                f"Error running work request Internal/workflow ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow(self) -> None:
        """A workflow is populated and left running."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def populate() -> None:
            wr.create_child_worker("noop")

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=populate,
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)

    def test_workflow_empty(self) -> None:
        """An empty workflow is populated and marked as completed."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate"
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.SUCCESS)

    @preserve_db_task_registry()
    def test_workflow_empty_sub_workflow(self) -> None:
        """A workflow with an empty sub-workflow isn't prematurely completed."""

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            TASK_NAME = "example"

            @override
            def populate(self) -> None:
                noop1 = self.work_request.create_child_workflow("noop")
                self.orchestrate_child(noop1)

                # Normally we'd orchestrate this sub-workflow as well, but
                # we skip that for the purposes of this test.
                self.work_request.create_child_workflow("noop")

        wr = self.playground.create_workflow(
            task_name="example", mark_running=True
        )

        self.assertTrue(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertQuerySetEqual(
            wr.children.order_by("id").values_list("status", "result"),
            [
                (WorkRequest.Statuses.COMPLETED, WorkRequest.Results.SUCCESS),
                (WorkRequest.Statuses.PENDING, WorkRequest.Results.NONE),
            ],
        )

    @preserve_db_task_registry()
    def test_workflow_computes_dynamic_data(self) -> None:
        """Dynamic task data is computed for workflows."""

        class ExampleDynamicData(BaseDynamicTaskData):
            dynamic: str

        class ExampleWorkflow(
            SampleWorkflow[BaseWorkflowData, ExampleDynamicData]
        ):
            TASK_NAME = "example"

            @override
            def compute_dynamic_data(self) -> ExampleDynamicData:
                return ExampleDynamicData(dynamic="foo")

            @override
            def populate(self) -> None:
                self.work_request.create_child_worker("noop")

        wr = self.playground.create_workflow(
            task_name="example", mark_running=True
        )

        self.assertTrue(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.dynamic_task_data, {"dynamic": "foo"})
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)

    def test_workflow_bad_task_data(self) -> None:
        """A workflow with bad task data is skipped."""
        # Bad task data would normally be caught by create_workflow.  Force
        # it to happen here by changing the task data after initial
        # creation.  (A more realistic case might be one where the
        # definition of a workflow changes but existing data isn't
        # migrated.)
        wr = self.playground.create_workflow(task_name="noop")
        wr.task_data = {"nonsense": ""}
        wr.configured_task_data = {"nonsense": ""}
        wr.save()
        self.playground.advance_work_request(wr, mark_running=True)
        expected_message = "Failed to configure: 1 validation error"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        assert isinstance(wr.output_data, OutputData)
        assert wr.output_data.errors is not None
        self.assertEqual(len(wr.output_data.errors), 1)
        self.assertTrue(
            wr.output_data.errors[0].message.startswith(expected_message)
        )
        self.assertEqual(wr.output_data.errors[0].code, "configure-failed")

    def test_workflows_with_failing_compute_dynamic_data(self) -> None:
        """A workflow where computing dynamic data fails is skipped."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        expected_message = "Failed to compute dynamic data: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow"
                ".compute_dynamic_data",
                side_effect=Exception("Boom"),
            ),
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="dynamic-data-failed"
                    )
                ]
            ),
        )

    def test_workflow_wrong_status(self) -> None:
        """A workflow in the wrong status is skipped."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_pending=False
        )
        expected_message = "Failed to mark work request as running"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate"
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_not_called()
        wr.refresh_from_db()
        # The work request remains blocked.
        self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)
        # No errors are logged in output_data, because the work request may
        # make progress in the future.
        self.assertIsNone(wr.output_data)

    def test_workflow_fails(self) -> None:
        """A workflow that fails is logged."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )
        expected_message = "Orchestrator failed: Boom"

        with (
            mock.patch(
                "debusine.server.workflows.noop.NoopWorkflow.populate",
                side_effect=ValueError("Boom"),
            ) as mock_noop_populate,
            self.assertLogsContains(
                f"Error running work request Workflow/noop ({wr.id}): "
                f"{expected_message}",
                logger="debusine.server.workflows.base",
                level=logging.WARNING,
            ),
        ):
            self.assertFalse(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="orchestrator-failed"
                    )
                ]
            ),
        )

    def test_workflow_unblocks_children(self) -> None:
        """Workflow children are unblocked if possible."""
        wr = self.playground.create_workflow(
            task_name="noop", mark_running=True
        )

        def populate() -> None:
            children = [wr.create_child_worker("noop") for _ in range(2)]
            children[1].add_dependency(children[0])

        with mock.patch(
            "debusine.server.workflows.noop.NoopWorkflow.populate",
            side_effect=populate,
        ) as mock_noop_populate:
            self.assertTrue(orchestrate_workflow(wr))

        mock_noop_populate.assert_called_once_with()
        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.RUNNING)
        self.assertEqual(wr.result, WorkRequest.Results.NONE)
        children = wr.children.order_by("id")
        self.assertEqual(children.count(), 2)
        self.assertEqual(children[0].status, WorkRequest.Statuses.PENDING)
        self.assertEqual(children[1].status, WorkRequest.Statuses.BLOCKED)

    def test_wrong_task_type(self) -> None:
        """Attempts to orchestrate non-workflows are logged."""
        wr = self.playground.create_worker_task(
            mark_pending=True, assign_new_worker=True
        )
        expected_message = "Does not have a workflow orchestrator"

        with self.assertLogsContains(
            f"Error running work request {wr.task_type}/{wr.task_name} "
            f"({wr.id}): {expected_message}",
            logger="debusine.server.workflows.base",
            level=logging.WARNING,
        ):
            self.assertFalse(orchestrate_workflow(wr))

        wr.refresh_from_db()
        self.assertEqual(wr.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(wr.result, WorkRequest.Results.ERROR)
        self.assertEqual(
            wr.output_data,
            OutputData(
                errors=[
                    OutputDataError(
                        message=expected_message, code="configure-failed"
                    )
                ]
            ),
        )


class TaskConfigurationTests(TestCase):
    """test :py:class:`Workflow` task configuration handling."""

    scenario = scenarios.DefaultContext()

    workflow_wr: ClassVar[WorkRequest]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.workflow_wr = cls.playground.create_workflow()

    @override
    def setUp(self) -> None:
        super().setUp()
        workflow = self.workflow_wr.get_task()
        assert isinstance(workflow, Workflow)
        self.workflow = workflow

    def test_work_request_ensure_child(self) -> None:
        package = self.playground.create_minimal_binary_package_artifact()
        for task_type, task_name, workflow_data, kwargs in (
            ("worker", "noop", WorkRequestWorkflowData(step="test"), {}),
            ("server", "servernoop", WorkRequestWorkflowData(step="test"), {}),
            ("internal", "workflow", WorkRequestWorkflowData(step="test"), {}),
            ("workflow", "noop", WorkRequestWorkflowData(step="test"), {}),
            ("signing", "noop", WorkRequestWorkflowData(step="test"), {}),
            (
                "wait",
                "externaldebsign",
                WorkRequestWorkflowData(step="test", needs_input=True),
                {"task_data": ExternalDebsignData(unsigned=package.pk)},
            ),
        ):
            with self.subTest(task_type=task_type, task_name=task_name):
                factory = partial(
                    getattr(
                        self.workflow, f"work_request_ensure_child_{task_type}"
                    ),
                    task_name=task_name,
                    workflow_data=workflow_data,
                    **kwargs,
                )
                wr, created = factory()
                self.assertIsNone(wr.configured_task_data)
                self.assertEqual(wr.status, WorkRequest.Statuses.BLOCKED)
                self.assertTrue(created)

                self.assertEqual(factory(), (wr, False))

                if task_type in {"worker", "internal", "workflow", "signing"}:
                    wr.mark_pending()
                    wr.mark_aborted()
                    with context.disable_permission_checks():
                        retried_wr = wr.retry()

                    self.assertEqual(factory(), (retried_wr, False))


class RTExampleWorkflowData(BaseWorkflowData):
    """Example Workflow Data for RuntimeParametersTests."""

    dynamic: str
    nested: dict[str, str]
    multiple: list[str]


class RuntimeParametersTests(TestCase):
    """Test Workflow.validate_runtime_parameters()."""

    workflow: ClassVar[
        type[Workflow[RTExampleWorkflowData, BaseDynamicTaskData]]
    ]

    @override
    @classmethod
    @preserve_db_task_registry()
    def setUpTestData(cls) -> None:
        """Add a sample Workflow to the test fixture."""
        super().setUpTestData()

        class RTExampleWorkflow(
            SampleWorkflow[RTExampleWorkflowData, BaseDynamicTaskData]
        ):
            TASK_NAME = "example"

        # https://github.com/python/mypy/issues/4717
        cls.workflow = RTExampleWorkflow  # type: ignore[type-abstract]

    def test_validate_any(self) -> None:
        self.workflow.validate_runtime_parameters(RuntimeParameter.ANY)

    def test_validate_empty_dict(self) -> None:
        self.workflow.validate_runtime_parameters({})

    def test_validate_parameter_any(self) -> None:
        self.workflow.validate_runtime_parameters(
            {"dynamic": RuntimeParameter.ANY}
        )

    def test_validate_parameter_values(self) -> None:
        self.workflow.validate_runtime_parameters({"dynamic": ["foo", "bar"]})

    def test_validate_parameter_values_wrong_type(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            r"dynamic is set to \[1\], not a list of valid values\.",
        ):
            self.workflow.validate_runtime_parameters({"dynamic": [1]})

    def test_validate_parameter_single_value(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            (
                r"dynamic is set to a string \('foo'\), not a list of valid "
                r"values\."
            ),
        ):
            self.workflow.validate_runtime_parameters({"dynamic": "foo"})

    def test_validate_unknown_parameter(self) -> None:
        with self.assertRaisesRegex(
            WorkflowValidationError,
            (
                r"unknown_parameter is not a task_data parameter for "
                r"RTExampleWorkflow"
            ),
        ):
            self.workflow.validate_runtime_parameters(
                {"unknown_parameter": RuntimeParameter.ANY}
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-client enrolling views."""

from typing import Any, override

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.core.exceptions import PermissionDenied
from django.db.models import QuerySet
from django.http import HttpRequest, HttpResponse, HttpResponseServerError
from django.shortcuts import get_object_or_404, redirect
from django.urls import reverse

from debusine.client.models import (
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
)
from debusine.db.models import Scope, Token
from debusine.db.models.auth import ClientEnroll
from debusine.web.views.base import BaseUIView, DetailViewBase


class ConfirmView(BaseUIView, DetailViewBase[ClientEnroll]):
    """Confirm a waiting client."""

    base_template = "web/_base_debusine.html"
    template_name = "web/enroll-confirm.html"
    title = "Confirm 'debusine setup'"
    model = ClientEnroll

    @override
    def get_object(
        self, queryset: QuerySet[ClientEnroll] | None = None
    ) -> ClientEnroll:
        """Return the ClientEnroll object to show."""
        assert queryset is None
        queryset = self.get_queryset()
        return get_object_or_404(queryset, nonce=self.kwargs["nonce"])

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Add the parsed payload."""
        ctx = super().get_context_data(**kwargs)
        payload: EnrollPayload | None
        try:
            payload = EnrollPayload.model_validate(self.object.payload)
        except ValueError:
            payload = None
        ctx["payload"] = payload
        if payload:
            ctx["scope_valid"] = Scope.objects.filter(
                name=payload.scope
            ).exists()
        return ctx

    @override
    def get(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        """Ask for a confirmation."""
        if not request.user.is_authenticated:
            return redirect(
                reverse("login") + "?next=" + request.get_full_path()
            )
        return super().get(request, *args, **kwargs)

    def post(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        """Provide a confirmation."""
        if not request.user.is_authenticated:
            raise PermissionDenied()
        self.object = self.get_object()
        try:
            payload = EnrollPayload.model_validate(self.object.payload)
        except ValueError:
            return HttpResponseServerError("Invalid payload in database")
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        channel_name = f"enroll.{payload.nonce}"

        try:
            outcome = EnrollOutcome(request.POST["outcome"])
        except ValueError:
            return HttpResponseServerError("Invalid outcome in POST data")

        if outcome == EnrollOutcome.CONFIRM:
            # Generate the new token
            token = Token.objects.create(
                token_type=Token.Types.USER,
                enabled=True,
                user=request.user,
                comment=(
                    f"obtained via 'debusine setup' on {payload.hostname}"
                ),
            ).key
        else:
            token = None

        confirmation = EnrollConfirmPayload(outcome=outcome, token=token)
        async_to_sync(channel_layer.send)(
            channel_name, confirmation.model_dump()
        )

        return redirect(
            reverse(
                "user:token-list", kwargs={"username": request.user.username}
            )
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for UI shortcuts."""

from typing import ClassVar, override

from django.http.request import HttpRequest
from django.template import Context
from django.utils.safestring import SafeString

from debusine.db.models import Artifact, WorkRequest
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views import ui_shortcuts
from debusine.web.views.base import Widget


class UIShortcutsTests(TestCase):
    """Tests for UI shortcuts."""

    scenario = scenarios.DefaultScopeUser()
    work_request: ClassVar[WorkRequest]
    source: ClassVar[Artifact]
    buildlog: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.work_request = cls.playground.create_worker_task()
        cls.source = cls.playground.create_source_artifact()
        cls.buildlog = cls.playground.create_build_log_artifact(
            work_request=cls.work_request
        )

    def _render(self, widget: Widget) -> str | SafeString:
        """Render a widget."""
        return widget.render(Context({"request": HttpRequest()}))

    def test_render(self) -> None:
        """Test rendering to HTML."""
        shortcut = ui_shortcuts.UIShortcut(
            label="LABEL", icon="ICON", url="URL"
        )
        self.assertEqual(
            self._render(shortcut),
            "<a class='btn btn-outline-secondary'"
            " href='URL' title='LABEL'>"
            "<span class='bi bi-ICON'></span>"
            "</a>",
        )

    def test_work_request_abort(self) -> None:
        """Test create_work_request_abort."""
        action = ui_shortcuts.create_work_request_abort(self.work_request)
        self.assertEqual(action.label, "Abort work request")
        self.assertEqual(action.icon, "x")
        self.assertEqual(action.url, self.work_request.get_absolute_url_abort())
        self.assertEqual(action.button_style, "danger")
        rendered = self._render(action)
        csrf_token = self.assertHTMLValid(rendered).xpath(
            '//input[@name="csrfmiddlewaretoken"]/@value'
        )[0]
        self.assertEqual(
            rendered,
            f"<form method='post'"
            f" action='{self.work_request.get_absolute_url_abort()}'>"
            f'<input type="hidden" name="csrfmiddlewaretoken"'
            f' value="{csrf_token}">'
            f"<button type='submit' class='btn btn-danger bi bi-x'"
            f" title='Abort work request'></button>"
            f"</form>",
        )

    def test_artifact_view(self) -> None:
        """Test create_artifact_view."""
        action = ui_shortcuts.create_artifact_view(self.source)
        self.assertEqual(action, ui_shortcuts.create_artifact_view(self.source))

        action = ui_shortcuts.create_artifact_view(self.buildlog)
        self.assertEqual(
            action, ui_shortcuts.create_artifact_view(self.buildlog)
        )

    def test_artifact_download(self) -> None:
        """Test create_artifact_download."""
        action = ui_shortcuts.create_artifact_download(self.source)
        self.assertEqual(action.label, "Download artifact")
        self.assertEqual(action.icon, "download")
        self.assertEqual(action.url, self.source.get_absolute_url_download())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Compute task status (workers, queue, etc.)."""

from typing import Any, override

from django.views.generic import TemplateView

from debusine.server.status import TaskQueueSummary, WorkerStatus
from debusine.tasks.models import WorkerType
from debusine.web.views.base import BaseUIView


class TaskStatusView(BaseUIView, TemplateView):
    """View to display task status."""

    base_template = "web/_base_debusine.html"
    template_name = "web/task_status.html"
    title = "Task queue"

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with arch_status."""
        context = super().get_context_data(**kwargs)

        status = WorkerStatus.get_status()

        # Re-oragnize status by worker type for table
        worker_status: dict[WorkerType, dict[str, Any]] = {}
        for worker_type, worker_summary in status.workers.items():
            worker_status[worker_type] = {
                "worker_tasks": status.worker_tasks[worker_type],
                "workers": worker_summary,
            }
        context["worker_status"] = worker_status

        # Re-organize status by architecture for table
        arch_status: dict[str, dict[str, Any]] = {}
        for arch, worker_summary in status.external_workers_arch.items():
            arch_status[arch] = {
                "worker_tasks_arch": status.worker_tasks_arch[arch],
                "external_workers_arch": worker_summary,
            }
        if status.worker_tasks_arch:
            # Sorts last
            arch_status["_none_"] = {
                "worker_tasks_arch": status.worker_tasks_arch.get(
                    None, TaskQueueSummary()
                ),
                "external_workers_arch": status.workers[WorkerType.EXTERNAL],
            }
        context["arch_status"] = arch_status

        return context



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""
Debusine integration tests.

Extract all work request debug logs to the specified path.
"""

import argparse
import sys
from collections.abc import Generator
from pathlib import Path
from typing import Any, cast

import yaml

sys.path.append(str(Path(__file__).parent.parent))

from utils.client import Client  # noqa: E402


def work_requests() -> Generator[dict[str, Any]]:
    """Return all work requests."""
    # `debusine work-request list` currently has no "unlimited" option, but
    # this should be enough in practice.
    result = Client.execute_command(
        "work-request", "list", "--yaml", "--limit", "10000"
    )
    try:
        data = yaml.safe_load(result.stdout)
        for item in data:
            yield cast(dict[str, Any], item)
    except Exception as exc:
        exc.add_note(f"Stdout: {result.stdout}")
        exc.add_note(f"Stderr: {result.stderr}")
        raise


def show_artifact(artifact_id: str | int) -> dict[str, Any]:
    """Return details for artifact_id."""
    result = Client.execute_command("artifact", "show", str(artifact_id))
    return cast(dict[str, Any], yaml.safe_load(result.stdout))


def extract_last_debug_logs_to(destdir: Path) -> None:
    """Extract the debug logs for the last work request to destdir."""
    for work_request in work_requests():
        for artifact_id in work_request["artifacts"]:
            artifact = show_artifact(artifact_id)
            if artifact["category"] == "debusine:work-request-debug-logs":
                artifact_dir = destdir / str(artifact_id)
                artifact_dir.mkdir(parents=True)
                Client.execute_command(
                    "artifact",
                    "download",
                    "--target-directory",
                    str(artifact_dir),
                    artifact_id,
                )


def main() -> None:
    """Parse arguments and run the job."""
    p = argparse.ArgumentParser()
    p.add_argument(
        "destdir", type=Path, help="Directory to extract artifacts into."
    )
    args = p.parse_args()

    extract_last_debug_logs_to(args.destdir)


if __name__ == "__main__":
    main()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for task models."""

import datetime as dt
import re
import textwrap
from typing import get_args
from unittest import TestCase

from pydantic import ValidationError

import debusine.tasks.models as task_models
from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
    RuntimeStatistics,
)
from debusine.tasks.models import build_lookup_string, parse_lookup_string


class LookupTests(TestCase):
    """Tests for collection item lookups."""

    def test_string(self) -> None:
        """Single-item lookups accept strings."""
        self.assertIsInstance(
            "debian@debian:archive",
            get_args(task_models.LookupSingle),
        )

    def test_integer(self) -> None:
        """Single-item lookups accept integers."""
        self.assertIsInstance(1, get_args(task_models.LookupSingle))

    def test_dict(self) -> None:
        """Multiple-item lookups accept dicts."""
        raw_lookup = {"collection": "debian@debian:archive"}
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump(),
            (
                {
                    "collection": "debian@debian:archive",
                    "child_type": "artifact",
                    "category": None,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_bad_type(self) -> None:
        with self.assertRaisesRegex(
            ValueError, r"Expected dict, got <class 'object'>"
        ):
            task_models.LookupDict.model_validate(object())

    def test_child_type(self) -> None:
        """`child_type` is accepted."""
        raw_lookup = {
            "collection": "debian@debian:archive",
            "child_type": "collection",
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump(),
            (
                {
                    "collection": "debian@debian:archive",
                    "child_type": "collection",
                    "category": None,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_category(self) -> None:
        """`category` is accepted."""
        raw_lookup = {
            "collection": "bookworm@debian:suite",
            "category": ArtifactCategory.SOURCE_PACKAGE,
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump(),
            (
                {
                    "collection": "bookworm@debian:suite",
                    "child_type": "artifact",
                    "category": ArtifactCategory.SOURCE_PACKAGE,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_name_matcher(self) -> None:
        """A `name` lookup is accepted."""
        raw_lookup = {"collection": "debian@debian:archive", "name": "foo"}
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump()[0]["name_matcher"],
            {
                "kind": task_models.CollectionItemMatcherKind.EXACT,
                "value": "foo",
            },
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_name_matcher_rejects_exact(self) -> None:
        """`name__exact` is rejected."""
        with self.assertRaisesRegex(
            ValueError, "Unrecognized matcher: name__exact"
        ):
            task_models.LookupMultiple.model_validate(
                {"collection": "debian@debian:archive", "name__exact": "foo"}
            )

    def test_name_matcher_lookup(self) -> None:
        """Lookups such as `name__contains` are accepted."""
        raw_lookup = {
            "collection": "debian@debian:archive",
            "name__contains": "foo",
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump()[0]["name_matcher"],
            {
                "kind": task_models.CollectionItemMatcherKind.CONTAINS,
                "value": "foo",
            },
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_name_matcher_rejects_extra_segments(self) -> None:
        """A `name` lookup followed by two segments is rejected."""
        with self.assertRaisesRegex(
            ValueError, "Unrecognized matcher: name__foo__bar"
        ):
            task_models.LookupMultiple.model_validate(
                {"collection": "debian@debian:archive", "name__foo__bar": "foo"}
            )

    def test_name_matcher_conflict(self) -> None:
        """Conflicting name lookups are rejected."""
        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                "Conflicting matchers: ['name', 'name__contains', "
                "'name__endswith', 'name__in', 'name__startswith']"
            ),
        ):
            task_models.LookupMultiple.model_validate(
                {
                    "collection": "debian@debian:archive",
                    "name": "foo",
                    "name__contains": "foo",
                    "name__endswith": "foo",
                    "name__in": ("foo", "bar"),
                    "name__startswith": "foo",
                }
            )

    def test_data_matcher(self) -> None:
        """A `data__KEY` lookup is accepted."""
        raw_lookup = {
            "collection": "debian@debian:archive",
            "data__package": "foo",
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump()[0]["data_matchers"],
            (
                (
                    "package",
                    {
                        "kind": task_models.CollectionItemMatcherKind.EXACT,
                        "value": "foo",
                    },
                ),
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_data_matcher_rejects_exact(self) -> None:
        """`data__KEY__exact` is rejected."""
        with self.assertRaisesRegex(
            ValueError, "Unrecognized matcher: data__package__exact"
        ):
            task_models.LookupMultiple.model_validate(
                {
                    "collection": "debian@debian:archive",
                    "data__package__exact": "foo",
                }
            )

    def test_data_matcher_lookup(self) -> None:
        """Lookups such as `data__KEY__contains` are accepted."""
        raw_lookup = {
            "collection": "debian@debian:archive",
            "data__package__contains": "foo",
            "data__version__startswith": "1.",
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump()[0]["data_matchers"],
            (
                (
                    "package",
                    {
                        "kind": task_models.CollectionItemMatcherKind.CONTAINS,
                        "value": "foo",
                    },
                ),
                (
                    "version",
                    {
                        "kind": (
                            task_models.CollectionItemMatcherKind.STARTSWITH
                        ),
                        "value": "1.",
                    },
                ),
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_data_matcher_rejects_extra_segments(self) -> None:
        """A `data` lookup followed by three segments is rejected."""
        with self.assertRaisesRegex(
            ValueError, "Unrecognized matcher: data__package__foo__bar"
        ):
            task_models.LookupMultiple.model_validate(
                {
                    "collection": "debian@debian:archive",
                    "data__package__foo__bar": "foo",
                }
            )

    def test_data_matcher_conflict(self) -> None:
        """Conflicting data lookups are rejected."""
        with self.assertRaisesRegex(
            ValueError,
            re.escape(
                "Conflicting matchers: ['data__package', "
                "'data__package__contains', 'data__package__endswith', "
                "'data__package__in', 'data__package__startswith']"
            ),
        ):
            task_models.LookupMultiple.model_validate(
                {
                    "collection": "debian@debian:archive",
                    "data__package": "foo",
                    "data__package__contains": "foo",
                    "data__package__endswith": "foo",
                    "data__package__in": ("foo", "bar"),
                    "data__package__startswith": "foo",
                }
            )

    def test_lookup_filter_single(self) -> None:
        """`lookup__KEY` with a single-lookup value is accepted."""
        subordinate_lookup = "single"
        raw_lookup = {
            "collection": "_@debian:package-build-logs",
            "lookup__foo": subordinate_lookup,
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        assert isinstance(lookup.root[0], task_models.LookupDict)
        self.assertEqual(
            lookup.model_dump()[0]["lookup_filters"],
            (("foo", subordinate_lookup),),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_lookup_filter_multiple(self) -> None:
        """`lookup__KEY` with a multiple-lookup value is accepted."""
        subordinate_lookup = ["internal@collections/name:build-amd64"]
        raw_lookup = {
            "collection": "_@debian:package-build-logs",
            "lookup__same_work_request": subordinate_lookup,
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        assert isinstance(lookup.root[0], task_models.LookupDict)
        self.assertEqual(
            lookup.root[0].lookup_filters,
            (
                (
                    "same_work_request",
                    task_models.LookupMultiple.model_validate(
                        subordinate_lookup
                    ),
                ),
            ),
        )
        self.assertEqual(
            lookup.model_dump()[0]["lookup_filters"],
            (("same_work_request", tuple(subordinate_lookup)),),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_error_on_empty(self) -> None:
        """`error_on_empty` is accepted."""
        raw_lookup = {
            "collection": "trixie@debian:suite",
            "error_on_empty": True,
        }
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump(),
            (
                {
                    "collection": "trixie@debian:suite",
                    "child_type": "artifact",
                    "category": None,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": True,
                },
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_list(self) -> None:
        """Multiple-item lookups accept lists."""
        raw_lookup = [
            {"collection": "debian@debian:archive"},
            {"collection": "kali@debian:archive"},
            "123@artifacts",
            124,
        ]
        lookup = task_models.LookupMultiple.model_validate(raw_lookup)
        self.assertEqual(
            lookup.model_dump(),
            (
                {
                    "collection": "debian@debian:archive",
                    "child_type": "artifact",
                    "category": None,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
                {
                    "collection": "kali@debian:archive",
                    "child_type": "artifact",
                    "category": None,
                    "name_matcher": None,
                    "data_matchers": (),
                    "lookup_filters": (),
                    "error_on_empty": False,
                },
                "123@artifacts",
                124,
            ),
        )
        self.assertEqual(lookup.export(), raw_lookup)

    def test_wrong_type(self) -> None:
        """Multiple-item lookups only accept dicts or lists."""
        with self.assertRaisesRegex(
            ValueError,
            "Lookup of multiple collection items must be a dictionary or a "
            "list",
        ):
            task_models.LookupMultiple.model_validate("foo")

    def test_parse_lookup_string_segments(self) -> None:
        segments, options = parse_lookup_string("a/b/c")
        self.assertEqual(segments, ["a", "b", "c"])
        self.assertIsNone(options)

    def test_parse_lookup_string_with_options(self) -> None:
        segments, options = parse_lookup_string("//opt1=val1:opt2=val2/a/b/c")
        self.assertEqual(segments, ["a", "b", "c"])
        self.assertEqual(options, {"opt1": "val1", "opt2": "val2"})

    def test_build_lookup_string(self) -> None:
        lookup = build_lookup_string("a", "b", "c")
        self.assertEqual(lookup, "a/b/c")

    def test_build_lookup_string_with_options(self) -> None:
        options = {"opt1": "val1", "opt2": "val2"}
        lookup = build_lookup_string("a", "b", "c", options=options)
        self.assertEqual(lookup, "//opt1=val1:opt2=val2/a/b/c")

    def test_round_trip_parse_build_lookup_string(self) -> None:
        lookup = "//opt1=val1:opt2=val2/a/b/c"
        segments, options = parse_lookup_string(lookup)
        lookup2 = build_lookup_string(*segments, options=options)
        self.assertEqual(lookup, lookup2)


class ActionRetryWithDelaysTests(TestCase):
    """Tests for :py:class:`ActionRetryWithDelays`."""

    def test_delays(self) -> None:
        """A valid `delays` field is accepted."""
        delays = ["30m", "1h", "2d", "1w"]
        action = task_models.ActionRetryWithDelays(delays=delays)
        self.assertEqual(
            action.model_dump(),
            {"action": "retry-with-delays", "delays": delays},
        )

    def test_delays_bad_format(self) -> None:
        """Items in `delays` must be integers followed by m/h/d/w."""
        with self.assertRaisesRegex(
            ValueError, "String should match pattern.*input_value='2 weeks'"
        ):
            task_models.ActionRetryWithDelays(delays=["2 weeks"])


class EnumTests(TestCase):
    """Tests for SystemImageBuild task data."""

    def test_enums_str(self) -> None:
        """Test enum stringification."""
        for enum_cls in (
            task_models.AutopkgtestNeedsInternet,
            task_models.BlhcFlags,
            task_models.DebDiffFlags,
            task_models.DebootstrapVariant,
            task_models.DiskImageFormat,
            task_models.LintianFailOnSeverity,
            task_models.MmDebstrapVariant,
            task_models.SbuildBuildComponent,
            task_models.SystemBootstrapRepositoryCheckSignatureWith,
            task_models.SystemBootstrapRepositoryType,
        ):
            with self.subTest(enum_cls=repr(enum_cls)):
                for el in enum_cls:
                    with self.subTest(el=el.value):
                        self.assertEqual(str(el), el.value)


class BaseDynamicTaskDataTests(TestCase):
    """Tests for ``BaseDynamicTaskData``."""

    sample = task_models.BaseDynamicTaskData(
        parameter_summary="ps",
        subject="s",
        runtime_context="rc",
        configuration_context="cc",
        task_input_fields={"field1": "value1"},
    )

    def test_serialize(self) -> None:
        serialized = self.sample.model_dump(mode="json")
        self.assertEqual(
            serialized,
            {
                "parameter_summary": "ps",
                "subject": "s",
                "runtime_context": "rc",
                "configuration_context": "cc",
                "task_input_fields": {"field1": "value1"},
            },
        )

        deserialized = task_models.BaseDynamicTaskData.model_validate(
            serialized
        )
        self.assertEqual(deserialized, self.sample)

    def test_serialize_exclude_unset(self) -> None:
        serialized = self.sample.model_dump(mode="json", exclude_unset=True)
        self.assertEqual(
            serialized,
            {
                "parameter_summary": "ps",
                "subject": "s",
                "runtime_context": "rc",
                "configuration_context": "cc",
                "task_input_fields": {"field1": "value1"},
            },
        )

        deserialized = task_models.BaseDynamicTaskData.model_validate(
            serialized
        )
        self.assertEqual(deserialized, self.sample)


class OutputDataTests(TestCase):
    """Tests for ``OutputData``."""

    def test_merge_disjoint(self) -> None:
        """``merge`` sets all the fields explicitly set on each model."""
        left = task_models.OutputData(
            runtime_statistics=RuntimeStatistics(duration=1)
        )
        right = task_models.OutputData(
            errors=[task_models.OutputDataError(message="message", code="code")]
        )

        merged = left.merge(right)

        self.assertEqual(
            merged.model_fields_set, {"runtime_statistics", "errors"}
        )
        self.assertEqual(
            merged.runtime_statistics, RuntimeStatistics(duration=1)
        )
        self.assertEqual(
            merged.errors,
            [task_models.OutputDataError(message="message", code="code")],
        )

    def test_merge_other_wins(self) -> None:
        """``merge`` prefers values from the ``other`` model."""
        left = task_models.OutputData(
            runtime_statistics=RuntimeStatistics(duration=1)
        )
        right = task_models.OutputData(
            runtime_statistics=RuntimeStatistics(duration=1, cpu_time=1)
        )

        merged = left.merge(right)

        self.assertEqual(merged.model_fields_set, {"runtime_statistics"})
        self.assertEqual(
            merged.runtime_statistics, RuntimeStatistics(duration=1, cpu_time=1)
        )


class LintianDataTests(TestCase):
    """Tests for Lintian task data."""

    def test_input_validation(self) -> None:
        """Test LintianInput validation."""
        task_models.LintianInput(source_artifact=1)
        task_models.LintianInput(
            binary_artifacts=task_models.LookupMultiple((1,))
        )
        task_models.LintianInput(
            binary_artifacts=task_models.LookupMultiple((1, 2))
        )

        error_msg = "One of source_artifact or binary_artifacts must be set"
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.LintianInput()
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.LintianInput(
                binary_artifacts=task_models.LookupMultiple(())
            )


class LintianDynamicDataTests(TestCase):
    """Tests for :py:class:`LintianDynamicData`."""

    def test_get_source_package_name(self) -> None:
        for subject in ("foo", None):
            with self.subTest(subject=subject):
                d = task_models.LintianDynamicData(
                    input_source_artifact_id=1,
                    input_binary_artifacts_ids=[1],
                    subject=subject,
                )
                self.assertEqual(d.get_source_package_name(), subject)


class DebDiffDataTests(TestCase):
    """Tests for DebDiff task data."""

    def test_input_validation(self) -> None:
        """Test DebDiffInput validation."""
        task_models.DebDiffInput(source_artifacts=[1, 2])
        task_models.DebDiffInput(
            binary_artifacts=[
                task_models.LookupMultiple.model_validate([1]),
                task_models.LookupMultiple.model_validate([2, 3]),
            ]
        )

        error_msg = (
            "Exactly one of source_artifacts or binary_artifacts must be set"
        )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput()
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput(
                source_artifacts=task_models.LookupMultiple.model_validate(
                    [1, 2]
                ),
                binary_artifacts=[
                    task_models.LookupMultiple.model_validate([1]),
                    task_models.LookupMultiple.model_validate([2]),
                ],
            )

        error_msg = "List should have at least 2 items after validation, not 1"
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput(source_artifacts=[1])
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput(
                binary_artifacts=[
                    task_models.LookupMultiple.model_validate([1])
                ]
            )

        error_msg = "List should have at most 2 items after validation, not 3"
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput(source_artifacts=[1, 2, 3])
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.DebDiffInput(
                binary_artifacts=[
                    task_models.LookupMultiple.model_validate([1]),
                    task_models.LookupMultiple.model_validate([2]),
                    task_models.LookupMultiple.model_validate([3]),
                ]
            )


class MakeSourcePackageUploadDynamicDataTests(TestCase):
    """Tests for :py:class:`MakeSourcePackageUploadDynamicData`."""

    def test_get_source_package_name(self) -> None:
        d = task_models.MakeSourcePackageUploadDynamicData(
            environment_id=1,
            input_source_artifact_id=1,
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), "foo")


class MergeUploadsDynamicDataTests(TestCase):
    """Tests for :py:class:`MergeUploadsDynamicData`."""

    def test_get_source_package_name(self) -> None:
        d = task_models.MergeUploadsDynamicData(
            environment_id=1,
            input_uploads_ids=[1],
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), "foo")


class SystemBootstrapDataTests(TestCase):
    """Tests for SystemBootstrap task data."""

    def test_repository_validation(self) -> None:
        """Test SystemBootstrapRepository validation."""
        common_kwargs = {
            "mirror": "https://deb.debian.org/deb",
            "suite": "bookworm",
        }

        task_models.SystemBootstrapRepository(**common_kwargs)

        error_msg = "List should have at least 1 item"
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(types=[], **common_kwargs)

        task_models.SystemBootstrapRepository(
            check_signature_with="external",
            keyring={"url": "https://example.com/keyring_file.txt"},
            **common_kwargs,
        )

        error_msg = (
            "repository requires 'keyring':"
            " 'check_signature_with' is set to 'external'"
        )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                check_signature_with="external", **common_kwargs
            )

    def test_repository_validation_duplicate_list_items(self) -> None:
        """Test SystemBootstrapRepository validation of duplicate list items."""
        common_kwargs = {
            "mirror": "https://deb.debian.org/deb",
            "suite": "bookworm",
        }

        task_models.SystemBootstrapRepository(**common_kwargs)

        error_msg = "contains duplicate items"
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                types=["deb", "deb"], **common_kwargs
            )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                components=["main", "main"], **common_kwargs
            )

    def test_keyring_file_url_under_usr_share_keyrings(self) -> None:
        """Keyring file:// URLs under /usr/share/keyrings/ are allowed."""
        task_models.SystemBootstrapRepository(
            mirror="https://deb.debian.org/deb",
            suite="bookworm",
            check_signature_with="external",
            keyring={
                "url": "file:///usr/share/keyrings/debian-archive-keyring.gpg"
            },
        )

    def test_keyring_file_url_under_usr_local_share_keyrings(self) -> None:
        """Keyring file:// URLs under /usr/local/share/keyrings/ are allowed."""
        task_models.SystemBootstrapRepository(
            mirror="https://deb.debian.org/deb",
            suite="bookworm",
            check_signature_with="external",
            keyring={
                "url": "file:///usr/local/share/keyrings/local-keyring.gpg"
            },
        )

    def test_keyring_file_url_not_under_allowed_directory(self) -> None:
        """Keyring file:// URLs not under an allowed directory are rejected."""
        common_kwargs = {
            "mirror": "https://deb.debian.org/deb",
            "suite": "bookworm",
            "check_signature_with": "external",
        }

        error_msg = (
            "file:// URLs for keyrings must be under /usr/share/keyrings/ or "
            "/usr/local/share/keyrings/"
        )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                keyring={"url": "file:///etc/passwd"}, **common_kwargs
            )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                keyring={"url": "file:///usr/share/keyringssuffix"},
                **common_kwargs,
            )
        with self.assertRaisesRegex(ValueError, error_msg):
            task_models.SystemBootstrapRepository(
                keyring={"url": "file:///usr/share/keyrings/../escape"},
                **common_kwargs,
            )


class AutopkgtestDataTests(TestCase):
    """Tests for Autopkgtest task data."""

    def test_validation(self) -> None:
        """Test AutopkgtestData validation."""
        common_kwargs = {
            "input": {
                "source_artifact": 1,
                "binary_artifacts": [1, 2],
            },
            "build_architecture": "amd64",
            "environment": 1,
        }

        task_models.AutopkgtestData(**common_kwargs)

        for field in ("global", "factor", "short", "install", "test", "copy"):
            with self.subTest(field=field):
                error_msg = (
                    rf"timeout\.{field}\n"
                    rf"  Input should be greater than or equal to 0"
                )
                with self.assertRaisesRegex(ValueError, error_msg):
                    task_models.AutopkgtestData(
                        timeout={field: -1}, **common_kwargs
                    )


class AutopkgtestDynamicDataTests(TestCase):
    """Tests for :py:class:`AutopkgtestDynamicData`."""

    def test_get_source_package_name(self) -> None:
        d = task_models.AutopkgtestDynamicData(
            environment_id=1,
            input_source_artifact_id=1,
            input_binary_artifacts_ids=[1],
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), "foo")


class BlhcDynamicDataTests(TestCase):
    """Tests for :py:class:`BlhcDynamicData`."""

    def test_get_source_package_name(self) -> None:
        d = task_models.BlhcDynamicData(
            environment_id=1,
            input_artifact_id=1,
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), "foo")


class ExtraExternalRepositoryTests(TestCase):
    """Tests for ExtraExternalRepository."""

    def test_valid_sources(self) -> None:
        """Accept a typical valid source."""
        task_models.ExtraExternalRepository(
            url="http://deb.debian.org/debian",
            suite="bookworm",
            components=["main", "non-free-firmware"],
        )

    def test_source_with_key(self) -> None:
        """Accept a valid flat source with a key."""
        task_models.ExtraExternalRepository(
            url="http://example.com/repo",
            suite="./",
            signing_key="PUBLIC KEY",
        )

    def test_flat_source_with_empty_components_list(self) -> None:
        """Rewrite an empty components list to None."""
        repo = task_models.ExtraExternalRepository(
            url="http://example.com/repo", suite="./", components=[]
        )
        self.assertIsNone(repo.components)

    def test_file_url(self) -> None:
        """Reject a file:// URL."""
        with self.assertRaisesRegex(
            ValueError, r"Input should be a valid URL, empty host"
        ):
            task_models.ExtraExternalRepository(
                url="file:/etc/passwd", suite="bookworm", components=["main"]
            )

    def test_crazy_suite(self) -> None:
        """Reject an unreasonable suite."""
        with self.assertRaisesRegex(
            ValueError, r"String should match pattern.*input_value='b@d_suite'"
        ):
            task_models.ExtraExternalRepository(
                url="http://example.com/",
                suite="b@d_suite",
                components=["main"],
            )

    def test_flat_repository_with_component(self) -> None:
        """Reject an flat repository with a component."""
        with self.assertRaisesRegex(
            ValueError, "Components cannot be specified for a flat repository"
        ):
            task_models.ExtraExternalRepository(
                url="http://example.com/", suite="flat/", components=["main"]
            )

    def test_non_flat_repository_without_component(self) -> None:
        """Reject a non-flat repository without a component."""
        with self.assertRaisesRegex(ValueError, "Components must be specified"):
            task_models.ExtraExternalRepository(
                url="http://example.com/", suite="bookworm"
            )

    def test_crazy_component(self) -> None:
        """Reject an unreasonable component."""
        with self.assertRaisesRegex(
            ValueError, r"String should match pattern.*input_value='foo/bar'"
        ):
            task_models.ExtraExternalRepository(
                url="http://example.com/",
                suite="bookworm",
                components=["foo/bar"],
            )

    def test_as_deb822_source_flat(self) -> None:
        """Render a flat sources.list entry to a string."""
        repo = task_models.ExtraExternalRepository(
            url="http://example.com/", suite="flat/"
        )
        self.assertEqual(
            repo.as_deb822_source(),
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.com/
                Suites: flat/
                """
            ),
        )

    def test_as_deb822_source_components(self) -> None:
        """Render a regular sources.list entry to a string."""
        repo = task_models.ExtraExternalRepository(
            url="http://example.com/",
            suite="bookworm",
            components=["foo", "bar"],
        )
        self.assertEqual(
            repo.as_deb822_source(),
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.com/
                Suites: bookworm
                Components: foo bar
                """
            ),
        )

    def test_as_deb822_source_signed(self) -> None:
        """Render a signed-by sources.list entry to a string."""
        repo = task_models.ExtraExternalRepository(
            url="http://example.com/",
            suite="bookworm",
            components=["foo", "bar"],
            signing_key="\n".join(
                (
                    "-----BEGIN PGP PUBLIC KEY BLOCK-----",
                    "",
                    "ABCDEFGHI",
                    "-----END PGP PUBLIC KEY BLOCK-----",
                )
            ),
        )
        self.assertEqual(
            repo.as_deb822_source(),
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.com/
                Suites: bookworm
                Components: foo bar
                Signed-By:
                 -----BEGIN PGP PUBLIC KEY BLOCK-----
                 .
                 ABCDEFGHI
                 -----END PGP PUBLIC KEY BLOCK-----
                """
            ),
        )
        self.assertEqual(
            repo.as_deb822_source(signed_by_filename="/signature.asc"),
            textwrap.dedent(
                """\
                Types: deb
                URIs: http://example.com/
                Suites: bookworm
                Components: foo bar
                Signed-By: /signature.asc
                """
            ),
        )


class SbuildDynamicDataTests(TestCase):
    """Tests for SbuildDynamicData."""

    def test_binnmu_maintainer_not_strict(self) -> None:
        """binnmu_maintainer does not have to be a deliverable email address."""
        task_models.SbuildDynamicData(
            input_source_artifact_id=1,
            binnmu_maintainer="Debusine <noreply@debusine-dev>",
        )

    def test_get_source_package_name(self) -> None:
        d = task_models.SbuildDynamicData(
            input_source_artifact_id=1,
            binnmu_maintainer="Debusine <noreply@debusine-dev>",
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), "foo")


class ImageCacheUsageLogEntryTests(TestCase):
    """Tests for ImageCacheUsageLogEntry."""

    def test_timestamp_validation_succeeds(self) -> None:
        """Test ImageCacheUsageLogEntry accepts aware timestamp."""
        task_models.ImageCacheUsageLogEntry(
            filename="foo", timestamp=dt.datetime.now(dt.UTC)
        )

    def test_timestamp_validation_fails(self) -> None:
        """Test ImageCacheUsageLogEntry rejects naive timestamp."""
        with self.assertRaisesRegex(ValueError, "timestamp is TZ-naive"):
            task_models.ImageCacheUsageLogEntry(
                filename="foo",
                timestamp=dt.datetime.now(),  # noqa: DTZ005
            )


class ImageCacheUsageLogTests(TestCase):
    """Tests for ImageCacheUsageLog."""

    def test_version_validation_succeeds(self) -> None:
        """Test ImageCacheUsageLog accepts version 1."""
        task_models.ImageCacheUsageLog(version=1)

    def test_version_validation_fails(self) -> None:
        """Test ImageCacheUsageLog rejects version 99."""
        with self.assertRaisesRegex(ValueError, "Unknown usage log version 99"):
            task_models.ImageCacheUsageLog(version=99)


class ExtractForSigningDataTests(TestCase):
    """Tests for ExtractForSigningData."""

    def test_backend_validation_succeeds(self) -> None:
        """Test ExtractForSigningData accepts no backend or backend=auto."""
        task_models.ExtractForSigningData(
            environment=1,
            input={"template_artifact": 2, "binary_artifacts": [3]},
        )
        task_models.ExtractForSigningData(
            environment=1,
            backend=task_models.BackendType.AUTO,
            input={"template_artifact": 2, "binary_artifacts": [3]},
        )

    def test_backend_validation_fails(self) -> None:
        """Test ExtractForSigningData rejects backends other than auto."""
        with self.assertRaisesRegex(
            ValueError,
            'ExtractForSigning only accepts backend "auto", not "qemu"',
        ):
            task_models.ExtractForSigningData(
                environment=1,
                backend=task_models.BackendType.QEMU,
                input={"template_artifact": 2, "binary_artifacts": [3]},
            )


class ExtractForSigningDynamicDataTests(TestCase):
    """Tests for :py:class:`ExtractForSigningDynamicData`."""

    def test_get_source_package_name(self) -> None:
        d = task_models.ExtractForSigningDynamicData(
            environment_id=1,
            input_template_artifact_id=1,
            input_binary_artifacts_ids=[1],
            subject="foo",
        )
        self.assertEqual(d.get_source_package_name(), None)


class AssembleSignedSourceDataTests(TestCase):
    """Tests for AssembleSignedSourceData."""

    def test_backend_validation_succeeds(self) -> None:
        """Test AssembleSignedSourceData accepts no backend or backend=auto."""
        task_models.AssembleSignedSourceData(
            environment=1, template=2, signed=[3]
        )
        task_models.AssembleSignedSourceData(
            environment=1,
            backend=task_models.BackendType.AUTO,
            template=2,
            signed=[3],
        )

    def test_backend_validation_fails(self) -> None:
        """Test AssembleSignedSourceData rejects backends other than auto."""
        with self.assertRaisesRegex(
            ValueError,
            'AssembleSignedSource only accepts backend "auto", not "qemu"',
        ):
            task_models.AssembleSignedSourceData(
                environment=1,
                backend=task_models.BackendType.QEMU,
                template=2,
                signed=[3],
            )


class AssembleSignedSourceDynamicDataTests(TestCase):
    """Tests for :py:class:`AssembleSignedSourceDynamicData."""

    def test_get_source_package_name(self) -> None:
        d = task_models.AssembleSignedSourceDynamicData(
            environment_id=1,
            template_id=1,
            signed_ids=[1],
            unsigned_ids=[1],
            subject="foo",
        )
        # In this case, subject is a binary package name, not a source package
        # name
        self.assertIsNone(d.get_source_package_name())


class EventReactionsTests(TestCase):
    """Tests for EventReactions."""

    def test_serialize_exclude_unset(self) -> None:
        update_collection_with_artifacts = (
            task_models.ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template="test",
                artifact_filters={"category": ArtifactCategory.TEST},
            )
        )
        value = task_models.EventReactions(
            on_success=[update_collection_with_artifacts]
        )
        self.assertEqual(
            value.model_dump(exclude_unset=True),
            {
                "on_success": [
                    update_collection_with_artifacts.model_dump(
                        exclude_unset=True
                    )
                ],
            },
        )

        # Try deserializing the serialized version
        value1 = task_models.EventReactions(
            **value.model_dump(exclude_unset=True)
        )
        self.assertEqual(value1, value)

    def test_add_reaction(self) -> None:
        update_collection_with_artifacts = (
            task_models.ActionUpdateCollectionWithArtifacts(
                collection="internal@collections",
                name_template="test",
                artifact_filters={"category": ArtifactCategory.TEST},
            )
        )

        value = task_models.EventReactions()
        value.add_reaction("on_success", update_collection_with_artifacts)

        self.assertEqual(
            value.model_dump(exclude_unset=True),
            {
                "on_success": [
                    update_collection_with_artifacts.model_dump(
                        exclude_unset=True
                    )
                ],
            },
        )

    def test_deserialize(self) -> None:
        """Test basic deserialization."""
        task_models.EventReactions.model_validate_json(
            """
            {
                "on_success": [
                    {
                        "action": "update-collection-with-artifacts",
                        "collection": "internal@collections",
                        "artifact_filters": {
                            "category": "debian:binary-package",
                            "data__deb_fields__Section": "python"
                        },
                        "name_template": "{package}_{version}",
                        "variables": {
                            "$package": "deb_fields.Package",
                            "$version": "deb_fields.Version"
                        }
                    }
                ],
                "on_failure": [
                    {
                        "action": "send-notification",
                        "channel": "admin-team",
                        "data": {
                            "cc": [ "qa-team@example.org " ],
                            "subject": "Work request ${work_request_id}"
                        }
                    },
                    {
                        "action": "send-notification",
                        "channel": "security-team"
                    }
                ]
            }
        """
        )


class DebDiffDynamicDataTests(TestCase):
    """Tests for DebDiffDynamicData."""

    def test_source_artifacts_and_binary_artifacts_set(self) -> None:
        with self.assertRaisesRegex(
            ValidationError,
            re.escape(
                "Only one of 'input_source_artifacts_ids' or "
                "'input_binary_artifacts_ids' may be set (not both)."
            ),
        ):
            task_models.DebDiffDynamicData(
                input_source_artifacts_ids=[1, 2],
                input_binary_artifacts_ids=[[3], [4]],
                environment_id=3,
            )

    def test_source_artifacts_and_binary_artifacts_none(self) -> None:
        # Does not raise
        task_models.DebDiffDynamicData(
            input_source_artifacts_ids=None,
            input_binary_artifacts_ids=None,
            environment_id=3,
        )

    def test_source_artifacts_only(self) -> None:
        task_models.DebDiffDynamicData(
            input_source_artifacts_ids=[1, 2], environment_id=3
        )

    def test_binary_artifacts_only(self) -> None:
        # Does not raise
        task_models.DebDiffDynamicData(
            input_binary_artifacts_ids=[[1, 2, 3], [10, 20, 30]],
            environment_id=3,
        )

    def test_get_source_package_name(self) -> None:
        for subject, expected in (
            ("foo", None),
            ("source:foo", "foo"),
            ("binary:foo", None),
        ):
            with self.subTest(subject=subject):
                d = task_models.DebDiffDynamicData(
                    environment_id=1,
                    subject=subject,
                )
                self.assertEqual(d.get_source_package_name(), expected)


class BaseTaskInputArtifactTests(TestCase):
    """Tests for :py:class:`BaseTaskInputArtifact`."""

    data = DebianBinaryPackage(
        srcpkg_name="unsigned",
        srcpkg_version="1.0",
        deb_fields={
            "Package": "hello",
            "Version": "1.0-1",
            "Architecture": "amd64",
        },
        deb_control_files=[],
    )

    def test_serialize_data_none(self) -> None:
        value = task_models.BaseTaskInputArtifact(
            artifact_id=1, category=ArtifactCategory.BINARY_PACKAGE, data=None
        )
        self.assertEqual(
            value.model_dump(),
            {
                "artifact_id": 1,
                "category": "debian:binary-package",
                "data": None,
            },
        )

    def test_serialize_category_none(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "Category must be present if data is supplied"
        ):
            task_models.BaseTaskInputArtifact(
                artifact_id=1, category=None, data=self.data
            )

    def test_serialize_invalid_category(self) -> None:
        with self.assertRaisesRegex(
            ValueError, "Category 'foo' is not supported"
        ):
            task_models.BaseTaskInputArtifact(
                artifact_id=1, category="foo", data=self.data
            )

    def test_serialize_category_data_none(self) -> None:
        value = task_models.BaseTaskInputArtifact(
            artifact_id=1, category=None, data=None
        )
        self.assertEqual(
            value.model_dump(),
            {"artifact_id": 1, "category": None, "data": None},
        )

    def test_deserialize_data_none(self) -> None:
        serialized = {
            "artifact_id": 1,
            "category": "debian:binary-package",
            "data": None,
        }
        deserialized = task_models.BaseTaskInputArtifact.model_validate(
            serialized
        )
        self.assertEqual(
            deserialized,
            task_models.BaseTaskInputArtifact(
                artifact_id=1,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=None,
            ),
        )

    def test_deserialize_category_none(self) -> None:
        serialized = {
            "artifact_id": 1,
            "category": None,
            "data": self.data,
        }
        with self.assertRaisesRegex(
            ValueError, "Category must be present if data is supplied"
        ):
            task_models.BaseTaskInputArtifact.model_validate(serialized)

    def test_deserialize_category_data_none(self) -> None:
        serialized = {"artifact_id": 1, "category": None, "data": None}
        deserialized = task_models.BaseTaskInputArtifact.model_validate(
            serialized
        )
        self.assertEqual(
            deserialized,
            task_models.BaseTaskInputArtifact(
                artifact_id=1,
                category=None,
                data=None,
            ),
        )


class InputArtifactSingleTests(TestCase):
    """Tests for :py:class:`InputArtifactSingle`."""

    value = task_models.InputArtifactSingle(
        lookup="test",
        label="field",
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )

    def test_serialize(self) -> None:
        assert self.value.data is not None
        self.assertEqual(
            self.value.model_dump(),
            {
                "lookup": "test",
                "label": "field",
                "artifact_id": 1,
                "category": "debian:binary-package",
                "data": self.value.data.model_dump(),
            },
        )

    def test_deserialize(self) -> None:
        serialized = self.value.model_dump()
        deserialized = task_models.InputArtifactSingle.model_validate(
            serialized
        )
        self.assertEqual(deserialized, self.value)


class InputArtifactMultipleTests(TestCase):
    """Tests for :py:class:`InputArtifactMultiple`."""

    value = task_models.InputArtifactMultiple(
        lookup=["test"],
        label="field",
        artifacts=[
            task_models.BaseTaskInputArtifact(
                artifact_id=1,
                category=ArtifactCategory.BINARY_PACKAGE,
                data=DebianBinaryPackage(
                    srcpkg_name="unsigned",
                    srcpkg_version="1.0",
                    deb_fields={
                        "Package": "hello",
                        "Version": "1.0-1",
                        "Architecture": "amd64",
                    },
                    deb_control_files=[],
                ),
            )
        ],
    )

    def test_serialize(self) -> None:
        self.assertEqual(
            self.value.model_dump(mode="json"),
            {
                "lookup": ["test"],
                "label": "field",
                "artifacts": [
                    self.value.artifacts[0].model_dump(),
                ],
            },
        )

    def test_deserialize(self) -> None:
        serialized = self.value.model_dump()
        deserialized = task_models.InputArtifactMultiple.model_validate(
            serialized
        )
        self.assertEqual(deserialized, self.value)


class InputCollectionSingleTests(TestCase):
    """Tests for :py:class:`InputCollectionSingle`."""

    value = task_models.InputCollectionSingle(
        lookup="test",
        label="field",
        collection_id=1,
        scope_name="scope",
        workspace_name="workspace",
        category=CollectionCategory.SUITE,
        name="name",
        data={"present": True},
    )

    def test_serialize(self) -> None:
        self.assertEqual(
            self.value.model_dump(mode="json"),
            {
                "lookup": "test",
                "label": "field",
                "collection_id": 1,
                "category": "debian:suite",
                "scope_name": "scope",
                "workspace_name": "workspace",
                "name": "name",
                "data": {"present": True},
            },
        )

    def test_deserialize(self) -> None:
        serialized = self.value.model_dump()
        deserialized = task_models.InputCollectionSingle.model_validate(
            serialized
        )
        self.assertEqual(deserialized, self.value)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""pytest fixtures and hooks for Debusine."""

import pytest

try:
    from pytest_django.lazy_django import django_settings_is_configured
except ImportError:  # integration tests

    def django_settings_is_configured() -> bool:
        """pytest_django has not configured Django."""
        return False


@pytest.hookimpl()
def pytest_unconfigure() -> None:  # pragma: no cover
    """Clean up any paths that we created."""
    if django_settings_is_configured():
        from debusine.django.django_utils import cleanup_temp_directory

        cleanup_temp_directory()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Task to assemble a signed source package."""

import json
import re
import shlex
import shutil
import subprocess
import tempfile
from collections.abc import Generator
from pathlib import Path, PurePath
from typing import Any, override

from debian.changelog import Changelog

from debusine.artifacts import (
    BinaryPackage,
    SigningOutputArtifact,
    SourcePackage,
    Upload,
)
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianBinaryPackage,
    WorkRequestResults,
    get_architectures,
    get_binary_package_name,
    get_binary_package_version,
)
from debusine.assets import KeyPurpose
from debusine.client.models import ArtifactResponse, RelationType
from debusine.tasks import BaseTaskWithExecutor, RunCommandTask
from debusine.tasks.inputs import (
    EnvironmentInput,
    MultiInput,
    SingleInput,
    Stage,
    UnsignedMultiInput,
)
from debusine.tasks.models import (
    AssembleSignedSourceData,
    AssembleSignedSourceDynamicData,
)
from debusine.utils import is_single_path_component, read_changes

# https://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-source
_re_package_name = re.compile(r"^[a-z0-9][a-z0-9+.-]+$")


class AssembleError(Exception):
    """An error occurred while assembling a signed source package."""


class AssembleSignedSource(
    RunCommandTask[AssembleSignedSourceData, AssembleSignedSourceDynamicData],
    BaseTaskWithExecutor[
        AssembleSignedSourceData, AssembleSignedSourceDynamicData
    ],
):
    """Task to assemble a signed source package."""

    TASK_VERSION = 1

    template = SingleInput(
        categories=[ArtifactCategory.BINARY_PACKAGE], stage=Stage.PENDING
    )
    environment = EnvironmentInput(stage=Stage.PENDING)
    signed = MultiInput()
    unsigned = UnsignedMultiInput(field="signed")

    @override
    def post_init(self) -> None:
        super().post_init()

        # Set by fetch_input.
        self._template_deb_name: str | None = None
        self._template_deb_file: Path | None = None
        self._signed_files: dict[tuple[KeyPurpose, str], Path] = {}
        self._built_using_ids: list[int] | None = None

        # Set by run.
        self._remote_execute_directory: PurePath | None = None
        self._local_execute_directory: Path | None = None
        self._source_package_artifact: SourcePackage | None = None
        self._upload_artifact: Upload | None = None

    @override
    def _cmdline(self) -> list[str]:
        """Unused abstract method from RunCommandTask."""
        raise NotImplementedError()

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        executor_available_key = f"executor:{self.backend}:available"
        return bool(worker_metadata.get(executor_available_key, False))

    @override
    def get_subject(self) -> str | None:
        assert isinstance(self.template.data, DebianBinaryPackage)
        return get_binary_package_name(self.template.data)

    @override
    def compute_dynamic_data(self) -> AssembleSignedSourceDynamicData:
        """
        Resolve artifact lookups for this task.

        :subject: binary package name out of ``template``
        :parameter_summary: binary package name and version (architecture)
        """
        template = self.template
        assert isinstance(template.data, DebianBinaryPackage)
        package_name = get_binary_package_name(template.data)
        package_version = get_binary_package_version(template.data)
        architectures = get_architectures(template.data)
        assert len(architectures) == 1
        package_architectures = architectures.pop()

        return AssembleSignedSourceDynamicData(
            environment_id=self.environment.artifact_id,
            template_id=template.artifact_id,
            signed_ids=self.signed.resolved_ids(),
            unsigned_ids=self.unsigned.resolved_ids(),
            subject=self.get_subject(),
            parameter_summary=f"{package_name}_{package_version} "
            f"({package_architectures})",
        )

    def _check_category(
        self, artifact: ArtifactResponse, expected_category: ArtifactCategory
    ) -> bool:
        if artifact.category != expected_category:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Expected template of category {expected_category}; got "
                    f"{artifact.category}"
                ],
            )
            return False
        return True

    def _get_related_artifact_ids(
        self,
        artifact_id: int,
        target_category: ArtifactCategory,
        relation_type: RelationType = RelationType.RELATES_TO,
    ) -> Generator[int]:
        assert self.debusine is not None

        for relation in self.debusine.relation_list(artifact_id=artifact_id):
            if relation.type == relation_type:
                artifact = self.debusine.artifact_get(
                    relation.target, workspace=self.workspace_name
                )
                if artifact.category == target_category:
                    yield relation.target

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Download the required artifacts."""
        assert self.dynamic_data

        template_dir = destination / "template-artifact"
        template_dir.mkdir()
        template_artifact = self.fetch_artifact(
            self.dynamic_data.template_id, template_dir
        )
        if not self._check_category(
            template_artifact, ArtifactCategory.BINARY_PACKAGE
        ):
            return False
        deb_files = [
            file for file in template_artifact.files if file.endswith(".deb")
        ]
        if len(deb_files) != 1:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Expected exactly one .deb package in template; got "
                    f"{deb_files}"
                ],
            )
            return False
        self._template_deb_name = BinaryPackage.create_data(
            template_artifact.data
        ).deb_fields["Package"]
        self._template_deb_file = template_dir / deb_files[0]

        for artifact_id in self.dynamic_data.signed_ids:
            signed_dir = destination / "signed-artifacts" / str(artifact_id)
            signed_dir.mkdir(parents=True)
            signed_artifact = self.fetch_artifact(artifact_id, signed_dir)
            if not self._check_category(
                signed_artifact, ArtifactCategory.SIGNING_OUTPUT
            ):
                return False
            signed_data = SigningOutputArtifact.create_data(
                signed_artifact.data
            )
            for result in signed_data.results:
                if result.output_file is not None:
                    self._signed_files[(signed_data.purpose, result.file)] = (
                        signed_dir / result.output_file
                    )

            self._built_using_ids = []
            for signing_input_id in self._get_related_artifact_ids(
                artifact_id, ArtifactCategory.SIGNING_INPUT
            ):
                for binary_package_id in self._get_related_artifact_ids(
                    signing_input_id, ArtifactCategory.BINARY_PACKAGE
                ):
                    self._built_using_ids.append(binary_package_id)

        return True

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """Configure task: ensure that the executor has dpkg-dev installed."""
        self._prepare_executor_instance()

        if self.executor_instance is None:
            raise AssertionError("self.executor_instance cannot be None")

        self.executor_instance.run(
            ["apt-get", "update"], run_as_root=True, check=True
        )
        self.executor_instance.run(
            ["apt-get", "--yes", "install", "dpkg-dev"],
            run_as_root=True,
            check=True,
        )

        return True

    def _run_cmd_or_raise(
        self, cmd: list[str], execute_directory: PurePath
    ) -> None:
        self.logger.info("Executing: %s", shlex.join(cmd))
        returncode = self.run_cmd(cmd, Path(execute_directory))
        self.logger.info("%s exited with code %s", cmd[0], returncode)
        if returncode != 0:
            raise AssembleError(f"{cmd[0]} exited with code {returncode}")

    def _pull_from_executor(self, source: PurePath, target: Path) -> None:
        """Pull a subdirectory from the executor, preserving timestamps."""
        assert self.executor_instance is not None
        assert self._remote_execute_directory is not None
        assert self._local_execute_directory is not None

        # Guard against accidentally passing
        # {local,remote}_execute_directory directly.
        assert source != self._remote_execute_directory
        assert target != self._local_execute_directory

        source_tar_path = PurePath(f"{source}.tar")
        target_tar_path = Path(f"{target}.tar")
        self._run_cmd_or_raise(
            ["tar", "-cf", str(source_tar_path), source.name], source.parent
        )
        self.executor_instance.file_pull(source_tar_path, target_tar_path)
        subprocess.run(
            ["tar", "--one-top-level", "-xf", target_tar_path],
            cwd=target.parent,
        )

    def _push_to_executor(self, source: Path, target: PurePath) -> None:
        """Push a subdirectory to the executor, preserving timestamps."""
        assert self.executor_instance is not None
        assert self._local_execute_directory is not None
        assert self._remote_execute_directory is not None

        # Guard against accidentally passing
        # {local,remote}_execute_directory directly.
        assert source != self._local_execute_directory
        assert target != self._remote_execute_directory

        source_tar_path = Path(f"{source}.tar")
        target_tar_path = PurePath(f"{target}.tar")
        subprocess.run(
            ["tar", "-cf", source_tar_path, source.name], cwd=source.parent
        )
        self.executor_instance.file_push(source_tar_path, target_tar_path)
        self._run_cmd_or_raise(
            ["tar", "--one-top-level", "-xf", str(target_tar_path)],
            target.parent,
        )

    def _extract_binary(self, deb: PurePath, target: PurePath) -> None:
        """Extract a binary package."""
        self._run_cmd_or_raise(
            ["dpkg-deb", "-x", str(deb), str(target)],
            target.parent,
        )

    def _check_template(self, code_signing_path: Path) -> None:
        """Check that the template is well-formed."""
        source_template_path = code_signing_path / "source-template"
        for file in (
            "debian/source/format",
            "debian/changelog",
            "debian/control",
            "debian/copyright",
            "debian/rules",
        ):
            if not (source_template_path / file).exists():
                raise AssembleError(
                    f"Required file {file} missing from template"
                )
        source_format = (
            (source_template_path / "debian" / "source" / "format")
            .read_text()
            .rstrip("\n")
        )
        if source_format != "3.0 (native)":
            raise AssembleError(
                f"Expected template source format '3.0 (native)'; "
                f"got '{source_format}'"
            )
        for file in ("debian/source/options", "debian/source/local-options"):
            if (source_template_path / file).exists():
                raise AssembleError(f"Template may not contain {file}")

    def _add_signed_files(
        self, code_signing_path: Path, source_path: Path
    ) -> None:
        """Add signed files to the source directory being assembled."""
        manifest = json.loads((code_signing_path / "files.json").read_text())
        for package, metadata in manifest["packages"].items():
            if _re_package_name.match(package) is None:
                raise AssembleError(f"'{package}' is not a valid package name")
            for file in metadata["files"]:
                file_path = PurePath(file["file"])
                if ".." in file_path.parts:
                    raise AssembleError(
                        f"File name '{file_path}' may not contain '..' segments"
                    )
                if file_path.is_absolute():
                    raise AssembleError(
                        f"File name '{file_path}' may not be absolute"
                    )
                purpose: KeyPurpose
                match file["sig_type"]:
                    case "efi":
                        purpose = KeyPurpose.UEFI
                    case _:
                        raise AssembleError(
                            f"Cannot handle sig_type '{file['sig_type']}'"
                        )
                qualified_file = f"{package}/{file['file']}"
                if (
                    signed_file := self._signed_files.get(
                        (purpose, qualified_file)
                    )
                ) is None:
                    raise AssembleError(
                        f"{purpose} signature of '{qualified_file}' not "
                        f"available"
                    )
                signature_path = (
                    source_path
                    / "debian"
                    / "signatures"
                    / package
                    / f"{file['file']}.sig"
                )
                signature_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy2(signed_file, signature_path)

    def _build_source_package(self, local_source_path: Path) -> PurePath:
        """
        Build a source package in the executor.

        Returns the path to the built `.changes` file in the executor.
        """
        assert self._remote_execute_directory is not None

        remote_source_path = self._remote_execute_directory / "source"
        self._push_to_executor(local_source_path, remote_source_path)
        self._run_cmd_or_raise(["dpkg-source", "-b", "."], remote_source_path)
        with open(local_source_path / "debian" / "changelog") as changelog_file:
            changelog = Changelog(changelog_file, max_blocks=1)
        remote_changes_path = (
            remote_source_path.parent
            / f"{changelog.package}_{changelog.version}_source.changes"
        )
        distribution = changelog.distributions.split()[0]
        self._run_cmd_or_raise(
            [
                "dpkg-genchanges",
                "-S",
                f"-DDistribution={distribution}",
                "-UCloses",
                f"-O{remote_changes_path}",
            ],
            remote_source_path,
        )
        return remote_changes_path

    def _create_artifacts(self, remote_changes_path: PurePath) -> None:
        """Create artifacts for the built source package."""
        assert self.executor_instance is not None
        assert self._local_execute_directory is not None

        local_changes_path = (
            self._local_execute_directory / remote_changes_path.name
        )
        self.executor_instance.file_pull(
            remote_changes_path, local_changes_path
        )
        changes = read_changes(local_changes_path)
        local_source_package_paths: list[Path] = []
        for checksum in changes["Files"]:
            # This was the result of running dpkg-genchanges ourselves, so
            # we can safely assert that it's reasonably well formed.  It's
            # also checked by read_changes.
            assert is_single_path_component(checksum["name"])
            local_source_package_paths.append(
                local_changes_path.parent / checksum["name"]
            )
            self.executor_instance.file_pull(
                remote_changes_path.parent / checksum["name"],
                local_source_package_paths[-1],
            )
        self._source_package_artifact = SourcePackage.create(
            name=changes["Source"],
            version=changes["Version"],
            files=local_source_package_paths,
        )
        self._upload_artifact = Upload.create(changes_file=local_changes_path)

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Do the main assembly work."""
        assert self._template_deb_name is not None
        assert self._template_deb_file is not None

        self._remote_execute_directory = PurePath(execute_directory)
        self._local_execute_directory = Path(
            tempfile.mkdtemp(prefix="debusine-assemble-signed-source-")
        )
        remote_template_path = self._remote_execute_directory / "template"
        local_template_path = self._local_execute_directory / "template"
        local_source_path = self._local_execute_directory / "source"

        # Running dpkg-deb on a user-provided file carries some risk, so we
        # do it within an executor, and then fetch the result using an
        # intermediate tar process.
        self._extract_binary(self._template_deb_file, remote_template_path)
        self._pull_from_executor(remote_template_path, local_template_path)
        local_code_signing_path = (
            local_template_path
            / "usr"
            / "share"
            / "code-signing"
            / self._template_deb_name
        )
        shutil.copytree(
            local_code_signing_path / "source-template", local_source_path
        )
        self._check_template(local_code_signing_path)
        self._add_signed_files(local_code_signing_path, local_source_path)
        changes_path = self._build_source_package(local_source_path)
        self._create_artifacts(changes_path)

        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,
    ) -> None:
        """Upload artifacts for the task."""
        assert self.work_request_id is not None
        assert self.debusine is not None

        if execution_result == WorkRequestResults.SUCCESS:
            assert self._built_using_ids is not None
            assert self._source_package_artifact is not None
            assert self._upload_artifact is not None

            uploaded_source_package_artifact = self.debusine.upload_artifact(
                self._source_package_artifact,
                workspace=self.workspace_name,
                work_request=self.work_request_id,
            )
            for built_using_id in self._built_using_ids:
                self.debusine.relation_create(
                    uploaded_source_package_artifact.id,
                    built_using_id,
                    RelationType.BUILT_USING,
                )
            uploaded_upload_artifact = self.debusine.upload_artifact(
                self._upload_artifact,
                workspace=self.workspace_name,
                work_request=self.work_request_id,
            )
            self.debusine.relation_create(
                uploaded_upload_artifact.id,
                uploaded_source_package_artifact.id,
                RelationType.EXTENDS,
            )

    @override
    def cleanup(self) -> None:
        """Clean up after running the task."""
        if self._local_execute_directory is not None:
            shutil.rmtree(self._local_execute_directory)

        super().cleanup()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Pagination support widgets."""

from functools import cached_property
from typing import Any, TYPE_CHECKING, TypeAlias, override

from django.core import paginator as django_paginator
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model

from debusine.web.views.base import TemplateWidget
from debusine.web.views.table.page_navigation import (
    PageNavigation,
    PageNavigationPreview,
    PageNavigationWidget,
)

if TYPE_CHECKING:
    from django.core.paginator import _SupportsPagination

    from debusine.web.views.table.table import Table

    PaginatorBase: TypeAlias = django_paginator.Paginator
    PageBase: TypeAlias = django_paginator.Page
    _SupportsPagination  # fake usage for vulture
else:
    # Django's Paginator doesn't support generic types at run-time yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    PaginatorBase = patch_runtime_generic_support(django_paginator.Paginator)
    PageBase = patch_runtime_generic_support(django_paginator.Page)


class TableHead[M: Model](TemplateWidget):
    """Render a standard table header."""

    template_name = "web/_table_header.html"

    def __init__(self, table: "Table[M]") -> None:
        """Build a TableHead widget."""
        self.table = table

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["table"] = self.table
        return ctx


class TableFoot[M: Model](TemplateWidget):
    """Render a standard table footer."""

    template_name = "web/_table_footer.html"

    def __init__(self, paginator: "Paginator[M]") -> None:
        """Build a TableFoot widget."""
        self.paginator = paginator

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["paginator"] = self.paginator
        ctx["table"] = self.paginator.table
        return ctx


class Paginator[M: Model](TemplateWidget, PaginatorBase[M]):
    """Pagination handling via widgets."""

    def __init__(
        self,
        table: "Table[M]",
        per_page: int,
        *,
        orphans: int = 0,
        allow_empty_first_page: bool = True,
    ) -> None:
        """
        Build the Pagination object.

        :param table: the table to paginate
        :param per_page: passed to Django's Paginator
        :param orphans: passed to Django's Paginator
        :param allow_empty_first_page: passed to Django's Paginator
        """
        super().__init__(
            table.rows,
            per_page,
            orphans=orphans,
            allow_empty_first_page=allow_empty_first_page,
        )
        self.table = table

    def __bool__(self) -> bool:
        """
        Check if the current page has information to display.

        Returns True if any of this is true:

         * The current page has rows to show
         * There is at least one active filter
        """
        return bool(self.page_obj) or bool(self.table.filters.active)

    @cached_property
    def page_obj(self) -> PageBase[M]:
        """Return the Django paginator.Page object."""
        return self.get_page(
            self.table.request.GET.get(f"{self.table.prefix}page")
        )

    @cached_property
    def page_navigation(self) -> PageNavigationWidget:
        """Return the PageNavigation widget."""
        if self.table.is_preview:
            return PageNavigationPreview(self)
        else:
            return PageNavigation(self)

    @cached_property
    def has_page_navigation(self) -> bool:
        """Check if page navigation is needed."""
        return self.num_pages > 1

    @cached_property
    def thead(self) -> TableHead[M]:
        """Return a standard <thead> widget."""
        return TableHead(self.table)

    @cached_property
    def tfoot(self) -> TableFoot[M]:
        """Return a standard <tfoot> widget."""
        return TableFoot(self)

    @override
    def get_template_name(self) -> str:
        if self.table.template_name is None:
            raise ImproperlyConfigured(
                "Table.template_name needs to be set"
                " to use Paginator as a widget"
            )
        return self.table.template_name

    @override
    def get_context_data(self) -> dict[str, Any]:
        ctx = super().get_context_data()
        ctx["paginator"] = self
        ctx["table"] = self.table
        return ctx

    @override
    def validate_number(self, number: Any) -> int:
        """Validate a 1-based page number, clamping values above last page."""
        try:
            if isinstance(number, float) and not number.is_integer():
                raise ValueError
            number = int(number)
        except (TypeError, ValueError):
            pass
        else:
            if number > self.num_pages:
                return self.num_pages

        return super().validate_number(number)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View for DebianPipeline work request."""

from typing import Any, override

from debusine.artifacts.models import TaskTypes
from debusine.db.models import WorkRequest
from debusine.db.tasks import DBTask
from debusine.server.workflows import DebianPipelineWorkflow, QAWorkflow
from debusine.server.workflows.models import RegressionTrackingUpdateQAResults
from debusine.web.views.qa import QATab
from debusine.web.views.work_request import Tab, WorkRequestPlugin


class DebianPipelineViewWorkRequestPlugin(WorkRequestPlugin):
    """View for DebianPipeline work request."""

    task_type = TaskTypes.WORKFLOW
    task_name = "debian_pipeline"
    task: DebianPipelineWorkflow

    def __init__(self, task: DBTask[Any]) -> None:
        """Find the relevant workflow elements."""
        super().__init__(task)

        self.tab_qa: QATab | None = self.make_tab_qa()

    def make_tab_qa(self) -> QATab | None:
        """Instantiate a QA tab, if QA data exist."""
        # Find QA sub-workflow
        try:
            qa_workflow = self.work_request.children.get(
                task_type=TaskTypes.WORKFLOW,
                task_name="qa",
                task_data__update_qa_results__in={
                    False,
                    RegressionTrackingUpdateQAResults.NO,
                },
            )
        except WorkRequest.DoesNotExist:
            return None

        qa_task = QAWorkflow(qa_workflow)

        tab = QATab(qa_task)
        if not tab.has_data:
            return None

        return tab

    @override
    def is_enabled(self) -> bool:
        return self.tab_qa is not None

    @override
    def add_tabs(self, tabs: list[Tab]) -> None:
        if self.tab_qa is not None:
            tabs.insert(1, self.tab_qa)

    @override
    def get_context_data(self) -> dict[str, Any]:
        return {}



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Supervise running tasks for the worker."""

import asyncio
import logging
import tempfile
import time as tm
import types
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from typing import Any, Literal, Self, assert_never

from debusine.artifacts.models import (
    RuntimeStatistics,
    TaskTypes,
    WorkRequestResults,
)
from debusine.client.debusine import Debusine
from debusine.client.models import WorkRequestResponse
from debusine.tasks import BaseExternalTask, TaskConfigError
from debusine.tasks.models import WorkerType
from debusine.worker.config import ConfigHandler
from debusine.worker.system_information import cpu_time

logger = logging.getLogger(__name__)


class TaskRunnerError(Exception):
    """Report an error while running a task."""

    def __init__(self, code: str, message: str, *args: Any) -> None:
        """
        Initialize a TaskRunnerError.

        :param code: the error code
        :param message: the error message, suitable for logging, with
          percent-expansion placeholders
        :param args: arguments to use to expand placeholders in ``message``.
        """
        self.code = code
        self.message = message
        self.args = args


class CollectStatistics:
    """Handle collecting statistics for a task execution."""

    def __init__(
        self,
        task: BaseExternalTask[Any, Any],
        statistics: RuntimeStatistics,
    ) -> None:
        """Initialize a CollectStatistics."""
        self.task = task

        # Statistics collected from the worker and task executors.
        self.statistics = statistics

        # The monotonic clock time when the task started executing.
        self.start_time: float = tm.monotonic()

        # The total CPU time used so far when the task started executing.
        # (We assume that the worker only runs one task at a time.)
        self.start_cpu_time: float = cpu_time()

        # Set this event when the task is done
        self.done = asyncio.Event()

        # Collection task
        self.collect_task: asyncio.Task[None] | None = None

    def gather(self) -> None:
        """Perform one round of statistics collection from the executor."""
        # We can only start collecting statistics once an executor is
        # instantiated.
        if self.task.executor is None:
            return

        # TODO: this stops the main loop. The executor API probably needs to
        #       have this made async
        statistics = self.task.executor.get_statistics(
            self.task.executor_instance
        )
        statistics.add_to(self.statistics)

    async def __aenter__(self) -> Self:
        """Start statistics collection."""
        self.gather()
        assert self.collect_task is None
        self.collect_task = asyncio.create_task(self.collect())
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Stop statistics collection."""
        self.done.set()
        assert self.collect_task is not None
        await self.collect_task
        self.collect_task = None

    async def collect(self) -> None:
        """Collect statistics until self.done is set."""
        try:
            done_task = asyncio.create_task(self.done.wait())
            while not self.done.is_set():
                self.gather()

                # Arbitrary polling interval.  We want to poll often enough to
                # get useful information about relatively brief spikes in disk
                # or RAM use, but not so often as to spend excessive amounts of
                # time gathering statistics.
                await asyncio.wait([done_task], timeout=10)
        finally:
            self.statistics.duration = int(tm.monotonic() - self.start_time)
            self.statistics.cpu_time = int(cpu_time() - self.start_cpu_time)


class TaskRunner:
    """Supervises running a task."""

    def __init__(
        self,
        config: ConfigHandler,
        worker_type: Literal[WorkerType.EXTERNAL, WorkerType.SIGNING],
        work_request: WorkRequestResponse,
        worker_workdir: Path,
    ) -> None:
        """
        Initialize a TaskRunner.

        :param config: worker configuration
        :param worker_type: worker type
        :param work_request: work request to run
        :param worker_workdir: ``Path`` to the directory where our working
          directory should be created
        """
        self.config = config
        self.worker_type = worker_type
        self.work_request = work_request
        self.workdir_tempdir = tempfile.TemporaryDirectory(
            prefix=f"WR-{self.work_request.id}-", dir=worker_workdir
        )
        self.workdir = Path(self.workdir_tempdir.name)
        self.workdir.chmod(0o755)
        self.client = Debusine(
            self.config.api_url,
            work_request.token,
            timeout=self.config.http_timeout,
            logger=logger,
            scope=work_request.scope,
        )
        # Running task
        self.task: BaseExternalTask[Any, Any] | None = None

        # Statistics collected from the worker and task executors.
        self.statistics = RuntimeStatistics()

    async def setup_task(self) -> None:
        """Instantiate the task to run."""
        task_type = TaskTypes(self.work_request.task_type)
        match self.worker_type:
            case WorkerType.EXTERNAL:
                expected_task_type = TaskTypes.WORKER
            case WorkerType.SIGNING:
                expected_task_type = TaskTypes.SIGNING
            case _ as unreachable:
                assert_never(unreachable)

        if task_type != expected_task_type:
            raise TaskRunnerError(
                "wrong-task-type",
                "Task: %s is of type %s, not %s",
                self.work_request.task_name,
                task_type,
                expected_task_type,
            )

        try:
            task_class = BaseExternalTask.class_from_name(
                TaskTypes(self.work_request.task_type),
                self.work_request.task_name,
            )
        except ValueError as exc:
            raise TaskRunnerError(
                "setup-failed",
                "Task: %s Error setup: %s",
                self.work_request.task_name,
                exc,
            )

        try:
            self.task = task_class(
                self.work_request.task_data, self.work_request.dynamic_task_data
            )
            # Inject the worker's configured HTTP timeout into the task instance
            self.task.http_timeout = self.config.http_timeout
        except TaskConfigError as exc:
            raise TaskRunnerError(
                "configure-failed",
                "Task: %s Error configure: %s",
                self.work_request.task_name,
                exc,
            )

        assert self.task.TASK_TYPE == self.work_request.task_type
        assert isinstance(self.task, BaseExternalTask)
        self.task.set_working_directory(self.workdir)
        self.task.set_worker_environment_variables(self.config.environment)

        self.task.configure_server_access(self.client)
        self.task.work_request_id = self.work_request.id
        self.task.workspace_name = self.work_request.workspace

    async def shutdown(self) -> None:
        """Shut down after execution."""
        try:
            self.workdir_tempdir.cleanup()
        except Exception as exc:
            logger.warning("Failed to remove '%s': %s", self.workdir, str(exc))

    async def __aenter__(self) -> Self:
        """Set up running the task."""
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: types.TracebackType | None,
    ) -> None:
        """Cleanup after running the task."""
        await self.shutdown()

    async def execute(self) -> WorkRequestResults:
        """
        Run the task.

        :returns: a completed Future with the task results or an exception
        :raises TaskRunError: on errors
        """
        assert self.task is not None
        # ThreadPool for BaseExternalTask.execute (e.g. Sbuild.exec).
        # We only allow a single thread.  We only run one task at a time
        # anyway, and signing workers need to run Django code in tasks which is
        # safest if done from only one thread.
        #
        # TODO: it is impossible to interrupt a running task using
        # concurrent.futures executors. Given that we're mostly executing
        # external commands, we can migrate to something based on
        # asyncio.Subprocess
        task_executor = ThreadPoolExecutor(max_workers=1)

        try:
            async with CollectStatistics(self.task, self.statistics):
                task_exec_future = task_executor.submit(
                    self.task.execute_logging_exceptions
                )
                try:
                    return await asyncio.wrap_future(task_exec_future)
                except Exception as exc:
                    raise TaskRunnerError(
                        "execute-failed",
                        "Task: %s Error execute: %s",
                        self.task.name,
                        exc,
                    )
        except asyncio.CancelledError:
            self.task.abort()
            raise
        finally:
            task_executor.shutdown(wait=False, cancel_futures=True)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for emitting metrics."""

import json
from typing import Literal, override
from unittest import TestCase as SimpleTestCase
from unittest import mock

import requests
import responses
from celery import shared_task, signals
from celery.contrib.testing.worker import start_worker
from celery.utils.nodenames import gethostname, nodename
from django.conf import settings
from django.test import override_settings
from django_prometheus.testutils import assert_metric_diff, save_registry
from prometheus_client import (
    CollectorRegistry,
    Counter,
    Gauge,
    Histogram,
    Summary,
)

from debusine.db.metrics import (
    BadMetricType,
    MetricNotFound,
    emit_metric,
    logger,
)
from debusine.project.celery import make_app
from debusine.test.django import TestCase, TransactionTestCase


class ExceptionTests(SimpleTestCase):
    """Test methods on our custom exceptions."""

    def test_MetricNotFound_str(self) -> None:
        e = MetricNotFound("some.metric")
        self.assertEqual(str(e), "Metric 'some.metric' not found")

    def test_BadMetricType_str(self) -> None:
        e = BadMetricType("some.metric", "Histogram")
        self.assertEqual(str(e), "Metric 'some.metric' is not a Histogram")


class EmitMetricDirectlyTests(TestCase):
    """Test :py:func:`emit_metric` when not running in a Celery task."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.registry = CollectorRegistry()
        self.prefix = self.__class__.__name__
        self.counter = Counter(
            f"{self.prefix}_test_counter",
            "test_counter",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.gauge = Gauge(
            f"{self.prefix}_test_gauge",
            "test_gauge",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.summary = Summary(
            f"{self.prefix}_test_summary",
            "test_summary",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.histogram = Histogram(
            f"{self.prefix}_test_histogram",
            "test_histogram",
            labelnames=("foo",),
            registry=self.registry,
        )
        self.enterContext(
            mock.patch("debusine.db.metrics.REGISTRY", new=self.registry)
        )
        self.frozen_registry = save_registry(registry=self.registry)

    def test_metric_not_found(self) -> None:
        with self.assertRaises(MetricNotFound) as raised:
            emit_metric(
                metric_type="counter", name="nonexistent", labels={}, value=1
            )
        self.assertEqual(raised.exception.name, "nonexistent")

    def test_bad_metric_type(self) -> None:
        metric_type: Literal["counter", "gauge", "summary", "histogram"]
        for metric_type, name in (
            ("counter", f"{self.prefix}_test_gauge"),
            ("gauge", f"{self.prefix}_test_counter"),
            ("summary", f"{self.prefix}_test_histogram"),
            ("histogram", f"{self.prefix}_test_summary"),
        ):
            with self.subTest(metric_type=metric_type):
                with self.assertRaises(BadMetricType) as raised:
                    emit_metric(
                        metric_type=metric_type, name=name, labels={}, value=1
                    )
                self.assertEqual(raised.exception.name, name)
                self.assertEqual(
                    raised.exception.metric_class_name, metric_type.capitalize()
                )

    def test_success(self) -> None:
        metric_type: Literal["counter", "gauge", "summary", "histogram"]
        for metric_type, name, value, expected_diffs in (
            (
                "counter",
                f"{self.prefix}_test_counter",
                1,
                ((f"{self.prefix}_test_counter_total", 1, {}),),
            ),
            (
                "gauge",
                f"{self.prefix}_test_gauge",
                2.5,
                ((f"{self.prefix}_test_gauge", 2.5, {}),),
            ),
            (
                "summary",
                f"{self.prefix}_test_summary",
                512,
                (
                    (f"{self.prefix}_test_summary_count", 1, {}),
                    (f"{self.prefix}_test_summary_sum", 512, {}),
                ),
            ),
            (
                "histogram",
                f"{self.prefix}_test_histogram",
                10.0,
                (
                    (f"{self.prefix}_test_histogram_bucket", 0, {"le": "7.5"}),
                    (f"{self.prefix}_test_histogram_bucket", 1, {"le": "10.0"}),
                    (f"{self.prefix}_test_histogram_count", 1, {}),
                    (f"{self.prefix}_test_histogram_sum", 10.0, {}),
                ),
            ),
        ):
            with self.subTest(metric_type=metric_type):
                emit_metric(
                    metric_type=metric_type,
                    name=name,
                    labels={"foo": "bar"},
                    value=value,
                )
                for (
                    expected_name,
                    expected_diff,
                    expected_labels,
                ) in expected_diffs:
                    with self.subTest(expected_name=expected_name):
                        assert_metric_diff(
                            self.frozen_registry,
                            expected_diff,
                            expected_name,
                            registry=self.registry,
                            foo="bar",
                            **expected_labels,
                        )


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task  # type: ignore[untyped-decorator]
def _emit_metric(
    metric_type: Literal["counter", "gauge", "summary", "histogram"],
    name: str,
    labels: dict[str, str],
    value: float,
) -> None:
    """Sample Celery task to emit a metric."""
    emit_metric(metric_type=metric_type, name=name, labels=labels, value=value)


@override_settings(CELERY_BROKER_URL="memory://")
class EmitMetricFromCeleryTaskTests(TransactionTestCase):
    """Test :py:func:`emit_metric` when running in a Celery task."""

    def test_success(self) -> None:
        url = f"https://{settings.DEBUSINE_FQDN}/api/1.0/open-metrics/emit/"

        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(
                make_app(),
                perform_ping_check=False,
                hostname=nodename("celery", gethostname()),
            ) as celery_worker,
        ):
            metric_type: Literal["counter", "gauge", "summary", "histogram"]
            for metric_type, name, value, _expected_diffs in (
                (
                    "counter",
                    "test_counter",
                    1,
                    (("test_counter_total", 1, {}),),
                ),
                ("gauge", "test_gauge", 2.5, (("test_gauge", 2.5, {}),)),
                (
                    "summary",
                    "test_summary",
                    512,
                    (
                        ("test_summary_count", 1, {}),
                        ("test_summary_sum", 512, {}),
                    ),
                ),
                (
                    "histogram",
                    "test_histogram",
                    10.0,
                    (
                        ("test_histogram_bucket", 0, {"le": "7.5"}),
                        ("test_histogram_bucket", 1, {"le": "10.0"}),
                        ("test_histogram_count", 1, {}),
                        ("test_histogram_sum", 10.0, {}),
                    ),
                ),
            ):
                with (
                    self.subTest(metric_type=metric_type),
                    responses.RequestsMock() as rsps,
                ):
                    rsps.add(
                        responses.POST, url, status=requests.codes.no_content
                    )
                    _emit_metric.delay(
                        metric_type=metric_type,
                        name=name,
                        labels={"foo": "bar"},
                        value=value,
                    ).get()
                    rsps.assert_call_count(url, 1)
                    assert rsps.calls[0].request.body is not None
                    self.assertEqual(
                        json.loads(rsps.calls[0].request.body),
                        {
                            "metric_type": metric_type,
                            "name": name,
                            "labels": {"foo": "bar"},
                            "value": value,
                        },
                    )

            signals.worker_shutdown.send(sender=celery_worker)

    def test_failure(self) -> None:
        with (
            mock.patch("debusine.django.django_utils.test_data_override"),
            start_worker(
                make_app(),
                perform_ping_check=False,
                hostname=nodename("celery", gethostname()),
            ) as celery_worker,
            responses.RequestsMock(),
            self.assertLogs(
                logger=logger,
            ) as logs,
        ):
            _emit_metric.delay(
                metric_type="counter",
                name="test_counter",
                labels={"foo": "bar"},
                value=1.0,
            ).get()
        signals.worker_shutdown.send(sender=celery_worker)
        self.assertRegex(
            logs.output[0],
            (
                r"Failed to report metric back to the server: .* "
                r"Error: Connection refused by Responses .*"
            ),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the enroll views."""

import datetime as dt
import secrets
from typing import override

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.test import Client
from django.urls import reverse
from rest_framework import status

from debusine.client.models import (
    EnrollConfirmPayload,
    EnrollOutcome,
    EnrollPayload,
)
from debusine.db.models import Token
from debusine.db.models.auth import ClientEnroll
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.tests.utils import ViewTestMixin


class ConfirmViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`ConfirmView`."""

    scenario = scenarios.DefaultScopeUser()
    payload: EnrollPayload
    url: str
    xpath_form = "//form[@id='confirm-form']"
    xpath_bad_scope = "//div[@id='bad-scope']"
    xpath_bad_payload = "//div[@id='bad-payload']"

    @override
    def setUp(self) -> None:
        super().setUp()
        # This is done in setUp to use a different nonce per test, making sure
        # that each test uses a different channel
        self.payload = EnrollPayload(
            nonce=secrets.token_urlsafe(8),
            challenge="correct horse battery staple",
            hostname="hostname",
            scope=self.scenario.scope.name,
        )
        self.url = reverse(
            "enroll:confirm", kwargs={"nonce": self.payload.nonce}
        )

    def test_GET_anonymous(self) -> None:
        """Anonymous GET redirects to login."""
        response = self.client.get(self.url)
        self.assertRedirects(response, reverse("login") + "?next=" + self.url)

    def test_POST_anonymous(self) -> None:
        """Anonymous GET redirects to login."""
        response = self.client.post(self.url, data=self.payload.model_dump())
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_GET_missing_nonce(self) -> None:
        """Invalid nonces on GET return 404."""
        self.client.force_login(self.scenario.user)
        response = self.client.get(self.url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_POST_missing_nonce(self) -> None:
        """Invalid nonces on POST return 404."""
        self.client.force_login(self.scenario.user)
        response = self.client.post(self.url)
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_GET_bad_payload(self) -> None:
        """Invalid payload in database on GET."""
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload={"mischief": True}
        )
        self.client.force_login(self.scenario.user)
        response = self.client.get(self.url)
        tree = self.assertResponseHTML(response)
        msg = self.assertHasElement(tree, self.xpath_bad_payload)
        self.assertTextContentEqual(
            msg,
            "The data filed for this enrollment seem to be corrupted:"
            " please retry the procedure with debusine setup.",
        )
        self.assertFalse(tree.xpath(self.xpath_form))

    def test_POST_bad_payload(self) -> None:
        """Invalid payload in database on POST."""
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload={"mischief": True}
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(self.url)
        self.assertEqual(
            response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        self.assertEqual(response.content, b"Invalid payload in database")

    def test_GET_bad_scope(self) -> None:
        """Invalid payload in database on GET."""
        self.payload.scope = "does-not-exist"
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload=self.payload.model_dump()
        )
        self.client.force_login(self.scenario.user)
        response = self.client.get(self.url)
        tree = self.assertResponseHTML(response)
        msg = self.assertHasElement(tree, self.xpath_bad_scope)
        self.assertTextContentEqual(
            msg,
            "The scope does-not-exist is not present"
            " in this Debusine instance.",
        )
        self.assertHasElement(tree, self.xpath_form)

    def test_POST_csrf(self) -> None:
        """POST enforces CSRF."""
        client = Client(enforce_csrf_checks=True)
        client.force_login(self.scenario.user)
        response = client.post(self.url)
        self.assertContains(
            response,
            "CSRF verification failed.",
            status_code=status.HTTP_403_FORBIDDEN,
        )

    def test_POST_bad_outcome(self) -> None:
        """Invalid outcome in POST data."""
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload=self.payload.model_dump()
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(self.url, data={"outcome": "mischief"})
        self.assertEqual(
            response.status_code, status.HTTP_500_INTERNAL_SERVER_ERROR
        )
        self.assertEqual(response.content, b"Invalid outcome in POST data")

    def test_valid_GET(self) -> None:
        """Test a valid GET."""
        ce = ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload=self.payload.model_dump()
        )
        ce.created_at = dt.datetime(2025, 1, 1, tzinfo=dt.UTC)
        ce.save()

        self.client.force_login(self.scenario.user)
        response = self.client.get(self.url)
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath(self.xpath_bad_payload))
        self.assertFalse(tree.xpath(self.xpath_bad_scope))
        p_created_at = self.assertHasElement(tree, "//p[@id='created-at']")
        self.assertRegex(
            self.get_node_text_normalized(p_created_at),
            r"The request was made at 2025-01-01 00:00 \(.+ ago\)"
            f" from hostname for scope {self.scenario.scope.name}.",
        )
        p_challenge = self.assertHasElement(tree, "//p[@id='challenge']")
        self.assertTextContentEqual(
            p_challenge,
            "Please check that the challenge words in debusine setup's"
            f" output are: {self.payload.challenge}.",
        )
        self.assertHasElement(tree, self.xpath_form)

    def test_valid_POST_confirm(self) -> None:
        """Test a valid POST with the confirm button."""
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload=self.payload.model_dump()
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(self.url, data={"outcome": "confirm"})

        # POST redirects to token list
        self.assertRedirects(
            response,
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            ),
        )

        # A confirmation message is sent via channels
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        channel_name = f"enroll.{self.payload.nonce}"
        msg = async_to_sync(channel_layer.receive)(channel_name)
        confirm_payload = EnrollConfirmPayload.model_validate(msg)

        self.assertEqual(confirm_payload.outcome, EnrollOutcome.CONFIRM)
        self.assertIsNotNone(confirm_payload.token)

        # Token is in the database
        assert confirm_payload.token is not None
        token = Token.objects.get(
            hash=Token._generate_hash(confirm_payload.token)
        )
        self.assertEqual(
            token.comment, "obtained via 'debusine setup' on hostname"
        )

    def test_valid_POST_cancel(self) -> None:
        """Test a valid POST with the cancel button."""
        initial_token_count = Token.objects.count()
        ClientEnroll.objects.create(
            nonce=self.payload.nonce, payload=self.payload.model_dump()
        )
        self.client.force_login(self.scenario.user)
        response = self.client.post(self.url, data={"outcome": "cancel"})

        # POST redirects to token list
        self.assertRedirects(
            response,
            reverse(
                "user:token-list",
                kwargs={"username": self.scenario.user.username},
            ),
        )

        # A confirmation message is sent via channels
        channel_layer = get_channel_layer()
        assert channel_layer is not None
        channel_name = f"enroll.{self.payload.nonce}"
        msg = async_to_sync(channel_layer.receive)(channel_name)
        confirm_payload = EnrollConfirmPayload.model_validate(msg)

        self.assertEqual(confirm_payload.outcome, EnrollOutcome.CANCEL)
        self.assertIsNone(confirm_payload.token)

        # No token was created
        self.assertEqual(Token.objects.count(), initial_token_count)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-signing command to generate a service private key.

This key is used to encrypt other private keys.
"""

from pathlib import Path
from typing import Any, override

from django.core.management import CommandParser
from nacl.public import PrivateKey

from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to generate a service private key."""

    help = "Generate a service private key"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        """Add CLI arguments for the generate_service_key command."""
        parser.add_argument(
            "output_file",
            type=Path,
            help="Write the new key to this file path.",
        )

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """Generate a new random key."""
        key = PrivateKey.generate()
        options["output_file"].touch(mode=0o600)
        options["output_file"].write_bytes(bytes(key))

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine websocket URLs Configuration."""

import os

from channels.routing import ProtocolTypeRouter, URLRouter
from django.core.asgi import get_asgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'debusine.project.settings')

django_asgi_app = get_asgi_application()


from debusine.server.middlewares.context import (  # noqa: E402
    ContextMiddlewareChannels,
)
from debusine.server.middlewares.token_last_seen_at import (  # noqa: E402
    TokenLastSeenAtMiddlewareChannels,
)
from debusine.server.routing import websocket_urlpatterns  # noqa: E402

application = ProtocolTypeRouter(
    {
        "http": django_asgi_app,
        "websocket": ContextMiddlewareChannels(
            TokenLastSeenAtMiddlewareChannels(
                # https://github.com/python/typeshed/pull/15080
                URLRouter(websocket_urlpatterns)  # type: ignore[arg-type]
            ),
        ),
    }
)



# Generated by Django 4.2.28 on 2026-05-07 15:25

import django.db.models.deletion
import pgtrigger.compiler
import pgtrigger.migrations
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0056_db_token_valid_type'),
    ]

    operations = [
        migrations.CreateModel(
            name='FileStoreJournal',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                ('size_change', models.BigIntegerField()),
            ],
        ),
        pgtrigger.migrations.RemoveTrigger(
            model_name='fileinstore',
            name='db_fileinstore_total_size_insert',
        ),
        pgtrigger.migrations.RemoveTrigger(
            model_name='fileinstore',
            name='db_fileinstore_total_size_delete',
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='fileinstore',
            trigger=pgtrigger.compiler.Trigger(
                name='db_fileinstore_total_size_insert',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='INSERT INTO db_filestorejournal (store_id, size_change) SELECT NEW.store_id, db_file.size FROM db_file WHERE db_file.id = NEW.file_id; RETURN NULL;',
                    hash='a516ca0f597081f2a76a685d1bd18d4135cac392',
                    operation='INSERT',
                    pgid='pgtrigger_db_fileinstore_total_size_insert_8cd7f',
                    table='db_fileinstore',
                    when='AFTER',
                ),
            ),
        ),
        pgtrigger.migrations.AddTrigger(
            model_name='fileinstore',
            trigger=pgtrigger.compiler.Trigger(
                name='db_fileinstore_total_size_delete',
                sql=pgtrigger.compiler.UpsertTriggerSql(
                    func='INSERT INTO db_filestorejournal (store_id, size_change) SELECT OLD.store_id, -db_file.size FROM db_file WHERE db_file.id = OLD.file_id; RETURN NULL;',
                    hash='c9a8d12ebaaf3f0cea64a0334b6f6a732588f2d3',
                    operation='DELETE',
                    pgid='pgtrigger_db_fileinstore_total_size_delete_07c98',
                    table='db_fileinstore',
                    when='AFTER',
                ),
            ),
        ),
        migrations.AddField(
            model_name='filestorejournal',
            name='store',
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name='journal_entries',
                to='db.filestore',
            ),
        ),
    ]
